from typing import (
    List,
    TypeVar,
)

from zuper_testint import TestContext, TestManagerInterface, tfor

import act4e_interfaces as I
from .data import filter_reqs, get_test_posets, purify_data
from .posets_utils import load_poset_tc, tm_load_poset, tm_save_poset

__all__: List[str] = []

X = TypeVar("X")


@tfor(I.FinitePosetRepresentation)
def test_FinitePosetRepresentation_basic(tm: TestManagerInterface) -> None:
    mks = tm.impof(I.FinitePosetRepresentation)
    posets = get_test_posets()

    for name, p in posets.items():
        r1 = tm_load_poset(tm, name, tm.store(purify_data(p.data)))
        r1_dumped = tm_save_poset(tm, name, r1)
        r2 = tm_load_poset(tm, name + "-re", r1_dumped)


@tfor(I.FinitePosetRepresentation, level="Product")
def test_FinitePosetRepresentation_Product(tm: TestManagerInterface) -> None:
    d = get_test_posets()
    use = filter_reqs(d, "poset_product")
    for k, v in use.items():
        tm.addtest(test_loading_poset, k, tid0=f"test_loading_poset_product-{k}")


def test_loading_poset(tc: TestContext, poset_name: str) -> None:
    poset = load_poset_tc(tc, poset_name)



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from .posets import FinitePoset, Poset
from .semigroups import FiniteMonoid, Monoid

__all__ = [
    "FiniteMonoidalPoset",
    "MonoidalPosetOperations",
    "MonoidalPosetOperations",
]

C = TypeVar("C")
E = TypeVar("E")


class MonoidalPoset(Generic[E], ABC):
    """Implementation of finite posets."""

    @abstractmethod
    def poset(self) -> Poset[E]:
        ...

    @abstractmethod
    def monoid(self) -> Monoid[E]:
        ...


class FiniteMonoidalPoset(Generic[E], MonoidalPoset[E], ABC):
    """Implementation of finite posets."""

    @abstractmethod
    def poset(self) -> FinitePoset[E]:
        ...

    @abstractmethod
    def monoid(self) -> FiniteMonoid[E]:
        ...


class MonoidalPosetOperations(ABC):
    """Implementation of finite posets."""

    @abstractmethod
    def is_monoidal_poset(self, fp: FinitePoset[E], fm: FiniteMonoid[E]) -> bool:
        """Check that the pair of poset and monoid make together a monoidal poset."""



from zuper_testint import TestManagerInterface, TestNotImplemented, tfor

import act4e_interfaces as I


@tfor(I.SetoidOperations)
def test_SetoidOperations(tm: TestManagerInterface) -> None:
    mks = tm.impof(I.SetoidOperations)
    raise TestNotImplemented()


@tfor(I.EnumerableSetsOperations)
def test_EnumerableSetsOperations(tm: TestManagerInterface) -> None:
    mks = tm.impof(I.EnumerableSetsOperations)
    raise TestNotImplemented()



import traceback
from typing import Any, Dict, TypeVar

import zuper_html as zh
from zuper_testint import TestContext, TestManagerInterface, TestRef, tfor

import act4e_interfaces as I
from .actual_tests import good_fsr
from .data import check_good_output, get_test_data, IOHelperImp, loadit_, purify_data, TestData
from .sets_utils import check_same_set, good_load_set, set_coherence

E = TypeVar("E")
X = TypeVar("X")


@tfor(I.FiniteSetRepresentation)
def test_FiniteSetRepresentation(tm: TestManagerInterface) -> None:
    d = get_test_data("set")
    res = {}
    for k, v in d.items():
        requires = v.requires
        # requires.pop('set_product', None)
        if requires:
            continue
        res[k] = v

    doit_sets(tm, res)


def doit_sets(tm: TestManagerInterface, collection2: Dict[str, TestData[I.FiniteSet_desc]]) -> None:
    def dump(tc: TestContext, fmr: I.FiniteSetRepresentation, fi: I.FiniteSet[Any]) -> I.FiniteSet_desc:
        h = IOHelperImp()
        res = fmr.save(h, fi)
        check_good_output(tc, res)
        return res

    def load(tc: TestContext, fmr: I.FiniteSetRepresentation, data: I.FiniteSet_desc) -> I.FiniteSet[Any]:
        h = IOHelperImp()
        return fmr.load(h, data)

    fsr = tm.impof(I.FiniteSetRepresentation)
    for setname, info in collection2.items():
        setdata = tm.store(purify_data(info.data))
        fs = tm_load_set(tm, setname, setdata)
        good_fs = good_load_set(setname)

        t1 = tm.addtest(set_coherence, fs)
        t2 = tm.addtest(check_same_set, fs, good_fs, extra_dep=t1)
        # your FSR, your set
        b = tm.addtest(dump, fsr, fs, tid0=f"set-{setname}-dump")
        tm.addtest(load, fsr, b, tid0=f"set-{setname}-load2")
        # our FSR, your set
        b = tm.addtest(dump, good_fsr, fs, tid0=f"set-{setname}-dump")
        tm.addtest(load, good_fsr, b, tid0=f"set-{setname}-load2")
        # your FSR, our set
        b = tm.addtest(dump, fsr, good_fs, tid0=f"set-{setname}-dump")
        tm.addtest(load, fsr, b, tid0=f"set-{setname}-load2")

    def check_throws(tc: TestContext, fsr_: I.FiniteSetRepresentation, data: I.FiniteSet_desc) -> None:
        h = IOHelperImp()
        try:
            res = fsr_.load(h, data)
        except NotImplementedError:
            raise
        except I.InvalidFormat:
            pass
        except BaseException as e:
            msg = zh.p(f"Invalid data should throw InvalidFormat, {type(e)} was thrown")
            tc.fail(msg, btr=traceback.format_exc(), data=repr(data))
        else:
            msg = zh.p(f"For invalid data you should throw InvalidFormat")
            tc.fail(msg, data=repr(data), returned=res)

    tm.addtest(check_throws, fsr, {})
    tm.addtest(check_throws, fsr, {"els": 1})
    tm.addtest(check_throws, fsr, [])
    tm.addtest(check_throws, fsr, "invalid")
    tm.addtest(check_throws, fsr, {"elements": 1})
    tm.addtest(check_throws, fsr, {"elements": {}})
    tm.addtest(check_throws, fsr, {"elements": "what ever"})


@tfor(I.FiniteSetRepresentation, level="Product")
def test_FiniteSetRepresentationProduct(tm: TestManagerInterface) -> None:
    d = get_test_data("set")
    res = {}
    for k, v in d.items():
        requires = dict(v.requires)
        requires.pop("set_product", None)
        if requires:
            continue
        res[k] = v

    doit_sets_simple(tm, res)


def doit_sets_simple(tm: TestManagerInterface, collection: Dict[str, TestData[I.FiniteSet_desc]]) -> None:
    def dump(tc: TestContext, fmr: I.FiniteSetRepresentation, fi: I.FiniteSet[Any]) -> I.FiniteSet_desc:
        h = IOHelperImp()
        res = fmr.save(h, fi)
        check_good_output(tc, res)
        return res

    def load(tc: TestContext, fmr: I.FiniteSetRepresentation, data: I.FiniteSet_desc) -> I.FiniteSet[Any]:
        h = IOHelperImp()
        return fmr.load(h, data)

    fsr = tm.impof(I.FiniteSetRepresentation)
    for setname, ss in collection.items():
        fs = tm_load_set(tm, setname, tm.store(purify_data(ss.data)))

        tm.addtest(set_coherence, fs)
        b = tm.addtest(dump, fsr, fs, tid0=f"set-{setname}-dump")
        tm.addtest(load, fsr, b, tid0=f"set-{setname}-load2")


def tm_load_set(
    tm: TestManagerInterface, name: str, data: TestRef[I.FiniteSet_desc]
) -> TestRef[I.FiniteSet[Any]]:
    h = IOHelperImp()

    # if 'data' in data:
    #     raise ValueError(data)
    def loadit(tc: TestContext, fsr: I.FiniteSetRepresentation, data1: I.FiniteSet_desc) -> I.FiniteSet[Any]:
        return loadit_(tc, fsr, h, data1, I.FiniteSet)

    fsr_ = tm.impof(I.FiniteSetRepresentation)
    return tm.addtest(loadit, fsr_, data, tid0=f"load-set-{name}")



from abc import ABC, abstractmethod
from typing import Any, Generic, Mapping, TypeVar

from .maps import FiniteMap
from .sets import FiniteSet, Setoid

__all__ = [
    "Semigroup",
    "FiniteSemigroup",
    "Monoid",
    "Group",
    "FiniteMonoid",
    "FiniteSemigroupConstruct",
    "FiniteFreeGroupConstruct",
    "FiniteGroup",
    "FreeSemigroup",
    "FiniteSemigroupMorphism",
    "GroupMorphism",
    "MonoidMorphism",
    "SemigroupMorphism",
    "FiniteGroupMorphism",
    "FiniteMonoidMorphism",
]

E = TypeVar("E")
C = TypeVar("C")


class Semigroup(Generic[E], ABC):
    @abstractmethod
    def carrier(self) -> Setoid[E]:
        ...

    @abstractmethod
    def compose(self, a: E, b: E) -> E:
        ...


class FiniteSemigroup(Generic[E], Semigroup[E], ABC):
    @abstractmethod
    def carrier(self) -> FiniteSet[E]:
        ...


class FreeSemigroup(Generic[C, E], Semigroup[E], ABC):
    @abstractmethod
    def unit(self, a: C) -> E:
        """From an element of the carrier, returns the element of the free semigroup"""


class FiniteSemigroupConstruct(ABC):
    @abstractmethod
    def free(self, fs: FiniteSet[C]) -> FreeSemigroup[C, Any]:
        """Construct the free semigroup on a set."""


class FreeGroup(Generic[C, E], FreeSemigroup[C, E], ABC):
    ...


class FiniteFreeGroupConstruct(ABC):
    @abstractmethod
    def free(self, fs: FiniteSet[C]) -> FreeGroup[C, E]:
        """Construct the free group on a set."""


class Monoid(Generic[E], Semigroup[E], ABC):
    @abstractmethod
    def identity(self) -> E:
        ...


class Group(Generic[E], Monoid[E], ABC):
    @abstractmethod
    def inverse(self, e: E) -> E:
        """Returns the inverse of an element"""


class FiniteMonoid(Generic[E], Monoid[E], FiniteSemigroup[E], ABC):
    ...


class FiniteGroup(Generic[E], Group[E], FiniteMonoid[E], ABC):
    ...


A = TypeVar("A")
B = TypeVar("B")


class SemigroupMorphism(Generic[A, B], ABC):
    @abstractmethod
    def source(self) -> Semigroup[A]:
        ...

    @abstractmethod
    def target(self) -> Semigroup[B]:
        ...

    @abstractmethod
    def mapping(self) -> Mapping[A, B]:
        ...


class MonoidMorphism(Generic[A, B], SemigroupMorphism[A, B], ABC):
    @abstractmethod
    def source(self) -> Monoid[A]:
        ...

    @abstractmethod
    def target(self) -> Monoid[B]:
        ...


class GroupMorphism(Generic[A, B], MonoidMorphism[A, B], ABC):
    @abstractmethod
    def source(self) -> Group[A]:
        ...

    @abstractmethod
    def target(self) -> Group[B]:
        ...


class FiniteSemigroupMorphism(Generic[A, B], SemigroupMorphism[A, B], ABC):
    @abstractmethod
    def source(self) -> FiniteSemigroup[A]:
        ...

    @abstractmethod
    def target(self) -> FiniteSemigroup[B]:
        ...

    @abstractmethod
    def mapping(self) -> FiniteMap[A, B]:
        ...


class FiniteMonoidMorphism(Generic[A, B], MonoidMorphism[A, B], ABC):
    @abstractmethod
    def source(self) -> FiniteMonoid[A]:
        ...

    @abstractmethod
    def target(self) -> FiniteMonoid[B]:
        ...

    @abstractmethod
    def mapping(self) -> FiniteMap[A, B]:
        ...


class FiniteGroupMorphism(Generic[A, B], GroupMorphism[A, B], ABC):
    @abstractmethod
    def source(self) -> FiniteGroup[A]:
        ...

    @abstractmethod
    def target(self) -> FiniteGroup[B]:
        ...

    @abstractmethod
    def mapping(self) -> FiniteMap[A, B]:
        ...



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from .categories import Category, FiniteCategory

__all__ = ["NaturalTransformation", "FiniteNaturalTransformation"]

Ob1 = TypeVar("Ob1")
Mor1 = TypeVar("Mor1")
Ob2 = TypeVar("Ob2")
Mor2 = TypeVar("Mor2")


class NaturalTransformation(Generic[Ob1, Mor1, Ob2, Mor2], ABC):
    @abstractmethod
    def cat1(self) -> Category[Ob1, Mor1]:
        ...

    @abstractmethod
    def cat2(self) -> Category[Ob2, Mor2]:
        ...

    @abstractmethod
    def component(self, ob: Ob1) -> Mor2:
        """Returns the component for a particular object in the first category.
        This is a morphism in the second category.
        """


class FiniteNaturalTransformation(
    Generic[Ob1, Mor1, Ob2, Mor2],
    NaturalTransformation[Ob1, Mor1, Ob2, Mor2],
    ABC,
):
    @abstractmethod
    def cat1(self) -> FiniteCategory[Ob1, Mor1]:
        ...

    @abstractmethod
    def cat2(self) -> FiniteCategory[Ob2, Mor2]:
        ...



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from .maps import FiniteMap
from .sets import FiniteSet, Setoid

__all__ = [
    "Relation",
    "FiniteRelation",
    "FiniteRelationOperations",
    "FiniteRelationCompose",
    "FiniteRelationProperties",
    "FiniteEndorelationProperties",
    "FiniteEndorelationOperations",
]

E = TypeVar("E")
E1 = TypeVar("E1")
E2 = TypeVar("E2")
E3 = TypeVar("E3")


class Relation(Generic[E1, E2], ABC):
    @abstractmethod
    def source(self) -> Setoid[E1]:
        ...

    @abstractmethod
    def target(self) -> Setoid[E2]:
        ...

    @abstractmethod
    def holds(self, e1: E1, e2: E2) -> bool:
        ...


class FiniteRelation(Generic[E1, E2], Relation[E1, E2], ABC):
    @abstractmethod
    def source(self) -> FiniteSet[E1]:
        ...

    @abstractmethod
    def target(self) -> FiniteSet[E2]:
        ...


class FiniteRelationProperties(ABC):
    @abstractmethod
    def is_surjective(self, fr: FiniteRelation[E1, E2]) -> bool:
        """Return True if the relation is surjective."""

    @abstractmethod
    def is_injective(self, fr: FiniteRelation[E1, E2]) -> bool:
        """Return True if the relation is injective."""

    @abstractmethod
    def is_defined_everywhere(self, fr: FiniteRelation[E1, E2]) -> bool:
        """Return True if the relation is defined everywhere."""

    @abstractmethod
    def is_single_valued(self, fr: FiniteRelation[E1, E2]) -> bool:
        """Return True if the relation is single-valued"""


class FiniteRelationCompose(ABC):
    @abstractmethod
    def compose(self, fr: FiniteRelation[E1, E2], fr2: FiniteRelation[E2, E3]) -> FiniteRelation[E1, E3]:
        """Compose two relations"""


class FiniteRelationOperations(ABC):
    @abstractmethod
    def transpose(self, fr: FiniteRelation[E1, E2]) -> FiniteRelation[E2, E1]:
        """Create the transposed of a relation"""

    @abstractmethod
    def as_relation(self, f: FiniteMap[E1, E2]) -> FiniteRelation[E1, E2]:
        """Re-writes a Finite Map as a relation."""


class FiniteEndorelationProperties(ABC):
    @abstractmethod
    def is_reflexive(self, fr: FiniteRelation[E, E]) -> bool:
        """Return True if the relation is reflexive."""

    @abstractmethod
    def is_irreflexive(self, fr: FiniteRelation[E, E]) -> bool:
        """Return True if the relation is irreflexive."""

    @abstractmethod
    def is_transitive(self, fr: FiniteRelation[E, E]) -> bool:
        """Return True if the relation is transitive."""

    @abstractmethod
    def is_symmetric(self, fr: FiniteRelation[E, E]) -> bool:
        """Return True if the relation is symmetric"""

    @abstractmethod
    def is_antisymmetric(self, fr: FiniteRelation[E, E]) -> bool:
        """Return True if the relation is antisymmetric"""

    @abstractmethod
    def is_asymmetric(self, fr: FiniteRelation[E, E]) -> bool:
        """Return True if the relation is asymmetric"""


class FiniteEndorelationOperations(ABC):
    @abstractmethod
    def transitive_closure(self, fr: FiniteRelation[E, E]) -> FiniteRelation[E, E]:
        """Returns the transitive closure of a relation"""



from typing import TypeVar

from zuper_testint import TestContext, TestManagerInterface, tfor

import act4e_interfaces as I
from .sets import tm_check_subset
from .sets_utils import good_load_set, set_coherence

E = TypeVar("E")
X = TypeVar("X")


@tfor(I.MakeSetUnion)
def test_MakeSetUnion(tm: TestManagerInterface) -> None:
    A = good_load_set("set1")
    B = good_load_set("set2")
    msu = tm.impof(I.MakeSetUnion)
    AuB = tm.addtest(set_union, msu, A, B)
    tm.addtest(set_coherence, AuB)
    tm_check_subset(tm, A, AuB, True, "A-union")
    tm_check_subset(tm, B, AuB, True, "B-union")


@tfor(I.MakeSetIntersection)
def test_MakeSetIntersection(tm: TestManagerInterface) -> None:
    A = good_load_set("set1")
    B = good_load_set("set2")
    msi = tm.impof(I.MakeSetIntersection)
    intersection = tm.addtest(set_intersection, msi, A, B)
    tm_check_subset(tm, intersection, A, True, "A-intersection")
    tm_check_subset(tm, intersection, B, True, "B-intersection")


def set_intersection(
    tc: TestContext, fso: I.MakeSetIntersection, a: I.FiniteSet[X], b: I.FiniteSet[X]
) -> I.FiniteSet[X]:
    return fso.intersection([a, b])  # FIXME: exceptions


def set_union(tc: TestContext, fso: I.MakeSetUnion, a: I.FiniteSet[X], b: I.FiniteSet[X]) -> I.FiniteSet[X]:
    return fso.union([a, b])  # FIXME: exceptions



import zuper_html as zh
from zuper_testint import find_imp, TestContext, tfor

import act4e_interfaces as I


@tfor(I.SimpleIntro)
def check_sum(tc: TestContext) -> None:
    si: I.SimpleIntro = find_imp(tc, I.SimpleIntro)
    a = 3
    b = 4
    tc.fail_not_equal2(a + b, si.sum(a, b), zh.span("Sum not correct"))



import itertools
from typing import Iterable, Iterator, Tuple, TypeVar

import zuper_html as zh
from zuper_testint import TestContext, TestManagerInterface, tfor

import act4e_interfaces as I
from .data import get_test_sets, purify_data
from .sets_representation import tm_load_set

E = TypeVar("E")
X = TypeVar("X")


@tfor(I.MakePowerSet)
def test_MakePowerSet(tm: TestManagerInterface) -> None:
    mps = tm.impof(I.MakePowerSet)

    d = get_test_sets()

    names = ["set_empty", "set_one", "set_two", "set2"]

    for name in names:
        info = d[name]
        # def get(p):
        #     return info.properties.get(p, None)

        # powerset = get('powerset')
        # if powerset:
        s1 = tm_load_set(tm, name, tm.store(purify_data(info.data)))
        # info2 = d[powerset]
        # s2 = tm_load_set(tm, powerset, tm.store(info2.data))

        tm.addtest(check_powerset, mps, s1, tid0=f"check_power_set-{name}")


def powerset(iterable: Iterable[X]) -> Iterator[Tuple[X, ...]]:
    "powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
    s = list(iterable)
    return itertools.chain.from_iterable(itertools.combinations(s, r) for r in range(len(s) + 1))


def check_powerset(tc: TestContext, fsp: I.MakePowerSet, a: I.FiniteSet[X]) -> None:
    if a.size() > 5:
        return

    res = tc.check_result(fsp, fsp.powerset, I.FiniteSetOfFiniteSubsets, a)

    S = list(a.elements())
    for _ in powerset(S):
        e = tc.check_result(res, res.construct, object, list(_))
        contained = tc.check_result(res, res.contains, bool, e)
        if not contained:
            tc.fail(
                zh.span("element not contained"), e=e
            )  # # logger.info(original=a, expected=b, res=str(res))  # check_set(c,  # res)  #
            # check_same_set(c, res, b)  # for e in res.elements():  #     inside = list(  # res.contents(
            # e))  #     e2 = res.construct(inside)  #     equal = res.equal(e2, e)  #     if  # not equal:
            #         tc.fail(zh.span('Elements should be equal'), e=e, inside=inside, e2=e2,  # equal=equal)



from abc import ABC, abstractmethod
from typing import Any, Generic, overload, Sequence, TypeVar

from .sets import FiniteSet, Setoid

C = TypeVar("C")  # composite
E = TypeVar("E")  # composite

__all__ = ["SetProduct", "FiniteSetProduct", "MakeSetProduct"]


class SetProduct(Generic[C, E], Setoid[E], ABC):
    """A set product is a setoid that can be factorized."""

    @abstractmethod
    def components(self) -> Sequence[Setoid[C]]:
        """Returns the components of the product"""

    @abstractmethod
    def pack(self, args: Sequence[C]) -> E:
        """Packs an element of each setoid into an element of the mapping"""
        raise NotImplementedError()

    @abstractmethod
    def unpack(self, args: E) -> Sequence[C]:
        """Packs an element of each setoid into an element of the mapping"""
        raise NotImplementedError()


class FiniteSetProduct(Generic[C, E], FiniteSet[E], SetProduct[C, E], ABC):
    """Specialization of SetProduct where we deal with FiniteSets"""

    @abstractmethod
    def components(self) -> Sequence[FiniteSet[C]]:
        """Returns the components"""


class MakeSetProduct(ABC):
    @overload
    def product(self, components: Sequence[FiniteSet[C]]) -> FiniteSetProduct[C, Any]:
        ...

    @overload
    def product(self, components: Sequence[Setoid[C]]) -> SetProduct[C, Any]:
        ...

    @abstractmethod
    def product(self, components: Sequence[Setoid[C]]) -> SetProduct[C, Any]:
        ...



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, List

from typing_extensions import TypedDict

from .helper import IOHelper
from .semigroups import FiniteGroup, FiniteMonoid, FiniteSemigroup
from .sets_representation import FiniteSet_desc
from .types import ConcreteRepr

__all__ = [
    "FiniteSemigroup_desc",
    "FiniteGroup_desc",
    "FiniteMonoid_desc",
    "FiniteGroupRepresentation",
    "FiniteMonoidRepresentation",
    "FiniteSemigroupRepresentation",
]


class FiniteSemigroup_desc(TypedDict):
    carrier: FiniteSet_desc
    composition: List[List[ConcreteRepr]]


class FiniteMonoid_desc(FiniteSemigroup_desc):
    neutral: ConcreteRepr


class FiniteGroup_desc(FiniteMonoid_desc):
    inverse: List[List[ConcreteRepr]]


class FiniteSemigroupRepresentation(ABC):
    @abstractmethod
    def load(self, h: IOHelper, s: FiniteSemigroup_desc) -> FiniteSemigroup[Any]:
        ...

    @abstractmethod
    def save(self, h: IOHelper, m: FiniteSemigroup[Any]) -> FiniteSemigroup_desc:
        ...


class FiniteMonoidRepresentation(ABC):
    @abstractmethod
    def load(self, h: IOHelper, s: FiniteMonoid_desc) -> FiniteMonoid[Any]:
        ...

    @abstractmethod
    def save(self, h: IOHelper, m: FiniteMonoid[Any]) -> FiniteMonoid_desc:
        ...


class FiniteGroupRepresentation(ABC):
    @abstractmethod
    def load(self, h: IOHelper, s: FiniteGroup_desc) -> FiniteGroup[Any]:
        ...

    @abstractmethod
    def save(self, h: IOHelper, m: FiniteGroup[Any]) -> FiniteGroup_desc:
        ...



from zuper_testint import TestManagerInterface, TestNotImplemented, tfor

import act4e_interfaces as I


@tfor(I.MonoidalPosetOperations)
def test_MonoidalPosetOperations(tm: TestManagerInterface) -> None:
    mks = tm.impof(I.MonoidalPosetOperations)
    raise TestNotImplemented()



from abc import ABC, abstractmethod
from typing import Any, List, Optional, TypeVar

from .posets import FinitePoset

__all__ = [
    "FinitePosetClosures",
    "FinitePosetMeasurement",
    "FinitePosetSubsetProperties",
    "FinitePosetSubsetProperties2",
    "FinitePosetInfSup",
    "FinitePosetMinMax",
]

C = TypeVar("C")
E = TypeVar("E")


class FinitePosetSubsetProperties(ABC):
    @abstractmethod
    def is_chain(self, fp: FinitePoset[E], s: List[E]) -> bool:
        """True if the given elements form a chain."""

    @abstractmethod
    def is_antichain(self, fp: FinitePoset[E], s: List[E]) -> bool:
        """True if the given elements form an antichain."""


class FinitePosetMeasurement(ABC):
    @abstractmethod
    def width(self, fp: FinitePoset[Any]) -> int:
        """Return the width of the poset."""

    @abstractmethod
    def height(self, fp: FinitePoset[Any]) -> int:
        """Return the height of the poset."""


class FinitePosetMinMax(ABC):
    @abstractmethod
    def minimal(self, fp: FinitePoset[E], S: List[E]) -> List[E]:
        """Return the minimal elements of S"""

    @abstractmethod
    def maximal(self, fp: FinitePoset[E], S: List[E]) -> List[E]:
        """Return the maximal elements of S"""


class FinitePosetInfSup(ABC):
    @abstractmethod
    def lower_bounds(self, fp: FinitePoset[E], s: List[E]) -> List[E]:
        """Computes the lower bounds for the subset (possibly empty)."""

    @abstractmethod
    def upper_bounds(self, fp: FinitePoset[E], s: List[E]) -> Optional[List[E]]:
        """Computes the upper bounds for the subset (possibly empty)."""

    @abstractmethod
    def infimum(self, fp: FinitePoset[E], s: List[E]) -> Optional[E]:
        """Computes the infimum / meet / greatest lower bound
        for the subset, or returns None if one does not exist."""

    @abstractmethod
    def supremum(self, fp: FinitePoset[E], s: List[E]) -> Optional[E]:
        """Computes the supremum for the subset if it exists,
        or returns None if one does not exist."""


class FinitePosetSubsetProperties2(ABC):
    @abstractmethod
    def is_lower_set(self, fp: FinitePoset[E], s: List[E]) -> bool:
        ...

    @abstractmethod
    def is_upper_set(self, fp: FinitePoset[E], s: List[E]) -> bool:
        ...


class FinitePosetClosures(ABC):
    @abstractmethod
    def upper_closure(self, fp: FinitePoset[E], s: List[E]) -> List[E]:
        ...

    @abstractmethod
    def lower_closure(self, fp: FinitePoset[E], s: List[E]) -> List[E]:
        ...



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Generic, Tuple, TypeVar

from .sets import FiniteSet, Setoid
from .sets_product import FiniteSetProduct, SetProduct
from .sets_sum import SetDisjointUnion

__all__ = [
    "FiniteCategory",
    "FiniteSemiCategory",
    "FiniteCategoryOperations",
    "SemiCategory",
    "SemiBiCategory",
    "Category",
    "CategoryOperations",
]

E1 = TypeVar("E1")
E2 = TypeVar("E2")

A = TypeVar("A")
B = TypeVar("B")
C = TypeVar("C")

O1 = TypeVar("O1")
O2 = TypeVar("O2")

Ob = TypeVar("Ob")
Mor = TypeVar("Mor")

Ob1 = TypeVar("Ob1")
Mor1 = TypeVar("Mor1")
Ob2 = TypeVar("Ob2")
Mor2 = TypeVar("Mor2")
Ob3 = TypeVar("Ob3")
Mor3 = TypeVar("Mor3")


class SemiBiCategory(Generic[Ob, Mor], ABC):
    @abstractmethod
    def objects(self) -> Setoid[Ob]:
        ...

    @abstractmethod
    def hom(self, ob1: Ob, ob2: Ob) -> Setoid[Mor]:
        ...

    @abstractmethod
    def compose(self, m1: Mor, m2: Mor) -> Mor:
        ...

    @abstractmethod
    def legs(self, m: Mor) -> Tuple[Ob, Ob]:
        """Return source and target of the Mor"""


class SemiCategory(Generic[Ob, Mor], SemiBiCategory[Ob, Mor], ABC):
    ...


class Category(Generic[Ob, Mor], SemiCategory[Ob, Mor], ABC):
    @abstractmethod
    def identity(self, ob: Ob) -> Mor:
        """Identity for the Ob"""


class FiniteSemiCategory(Generic[Ob, Mor], SemiCategory[Ob, Mor], ABC):
    @abstractmethod
    def objects(self) -> FiniteSet[Ob]:
        ...

    @abstractmethod
    def hom(self, ob1: Ob, ob2: Ob) -> FiniteSet[Mor]:
        ...


class FiniteCategory(Generic[Ob, Mor], FiniteSemiCategory[Ob, Mor], Category[Ob, Mor], ABC):
    ...


class CategoryOperations:
    @abstractmethod
    def product(
        self, c1: Category[Ob, Mor], c2: Category[Ob, Mor]
    ) -> Category[SetProduct[Ob, Any], SetProduct[Mor, Any]]:
        """Product of two categories."""

    @abstractmethod
    def disjoint_union(
        self, c1: Category[Ob, Mor], c2: Category[Ob, Mor]
    ) -> Category[SetDisjointUnion[Ob], Any]:  # TODO: better types
        """Disjoint union for the categories"""

    @abstractmethod
    def arrow(self, c1: Category[Ob, Mor]) -> Category[Mor, Any]:  # TODO: better types
        """Computes the arrow category"""

    @abstractmethod
    def twisted(self, c1: Category[Ob, Mor]) -> Category[Mor, Any]:  # TODO: better types
        """Computes the twisted arrow category"""


class FiniteCategoryOperations:
    @abstractmethod
    def product(
        self, c1: Category[Ob, Mor], c2: Category[Ob, Mor]
    ) -> Category[FiniteSetProduct[Any, Ob], FiniteSetProduct[Any, Mor]]:
        """Product of two categories."""

    @abstractmethod
    def disjoint_union(
        self, c1: FiniteCategory[Ob, Mor], c2: FiniteCategory[Ob, Mor]
    ) -> FiniteCategory[Any, Any]:  # TODO: better types

        """Disjoint union for the categories"""

    @abstractmethod
    def arrow(self, c1: FiniteCategory[Ob, Mor]) -> FiniteCategory[Any, Any]:  # TODO: better types
        """Computes the arrow category"""

    @abstractmethod
    def twisted_arrow(self, c1: FiniteCategory[Ob, Mor]) -> FiniteCategory[Any, Any]:  # TODO: better types
        """Computes the twisted arrow category"""



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from .posets import Poset
from .sets import FiniteSet

__all__ = [
    "MeetSemilattice",
    "Lattice",
    "JoinSemilattice",
    "FiniteLattice",
]

C = TypeVar("C")
E = TypeVar("E")


class MeetSemilattice(Generic[E], Poset[E], ABC):
    @abstractmethod
    def meet(self, x: E, y: E) -> E:
        ...

    @abstractmethod
    def top(self) -> E:
        ...


class JoinSemilattice(Generic[E], Poset[E], ABC):
    @abstractmethod
    def join(self, x: E, y: E) -> E:
        ...

    @abstractmethod
    def bottom(self) -> E:
        ...


class Lattice(Generic[E], JoinSemilattice[E], MeetSemilattice[E], ABC):
    ...


class FiniteLattice(Generic[E], Lattice[E], ABC):
    @abstractmethod
    def carrier(self) -> FiniteSet[E]:
        ...



from abc import ABC
from typing import Generic, TypeVar, Any

from .categories import FiniteCategory, Category
from .posets import Poset
from .dps import DP

Object = TypeVar("Object")
Morphism = TypeVar("Morphism")

__all__ = [
    "Profunctor",
    "FiniteProfunctor",
    "FiniteProfunctorOperations",
    "FiniteEnrichedCategory",
    "DPCategory",
]


class Profunctor(ABC):
    pass  # # def source(self) -> Category:  #     ...  #  # def target(self) -> Category:  #     ...  #
    # def functor(self) -> Functor:  #     ...


class FiniteProfunctor(ABC):
    pass  # # def cat1(self) -> FiniteCategory:  #     ...  #  # def cat2(self) -> FiniteCategory:  #
    # ...  #  # def functor(self) -> FiniteFunctor:  #     ...


class FiniteProfunctorOperations(ABC):
    pass  # # @abstractmethod  # def series(self, p1: FiniteProfunctor, p2: FiniteProfunctor) ->
    # FiniteProfunctor:  #     ...


class FiniteEnrichedCategory(Generic[Object, Morphism], FiniteCategory[Object, Morphism], ABC):
    pass  # def enrichment(self) -> FiniteMonoidalCategory:  #     ...


class DPCategory(Category[Poset[Any], DP[Any, Any]], ABC):
    ...



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from .relations import Relation
from .sets import (
    FiniteSet,
    Setoid,
)

__all__ = [
    "Poset",
    "FinitePoset",
]

C = TypeVar("C")
E = TypeVar("E")


class Poset(Generic[E], Relation[E, E], ABC):
    def source(self) -> Setoid[E]:
        return self.carrier()

    def target(self) -> Setoid[E]:
        return self.carrier()

    @abstractmethod
    def carrier(self) -> Setoid[E]:
        ...

    @abstractmethod
    def holds(self, a: E, b: E) -> bool:
        ...


class FinitePoset(Generic[E], Poset[E], ABC):
    @abstractmethod
    def carrier(self) -> FiniteSet[E]:
        ...



from abc import ABC, abstractmethod
from typing import Generic, TypeVar
from .categories import Category, FiniteCategory


__all__ = ["Functor", "FiniteFunctor"]


Ob1 = TypeVar("Ob1")
Mor1 = TypeVar("Mor1")
Ob2 = TypeVar("Ob2")
Mor2 = TypeVar("Mor2")
Ob3 = TypeVar("Ob3")
Mor3 = TypeVar("Mor3")


class Functor(Generic[Ob1, Mor1, Ob2, Mor2], ABC):
    @abstractmethod
    def source(self) -> Category[Ob1, Mor1]:
        ...

    @abstractmethod
    def target(self) -> Category[Ob2, Mor2]:
        ...

    @abstractmethod
    def f_ob(self, ob: Ob1) -> Ob2:
        """Effect on objects"""

    @abstractmethod
    def f_mor(self, m: Mor1) -> Mor2:
        """Effect on morphisms"""


class FiniteFunctor(Generic[Ob1, Mor1, Ob2, Mor2], Functor[Ob1, Mor1, Ob2, Mor2], ABC):
    @abstractmethod
    def source(self) -> FiniteCategory[Ob1, Mor1]:
        ...

    @abstractmethod
    def target(self) -> FiniteCategory[Ob2, Mor2]:
        ...


# TODO: functor composition



from typing import Any, TypeVar

import zuper_html as zh
from zuper_testint import find_imp, TestContext, TestManagerInterface, tfor

import act4e_interfaces as I
from .sets_utils import check_same_element, check_same_set, load_set_tc

E = TypeVar("E")
X = TypeVar("X")


@tfor(I.MakeSetDisjointUnion)
def test_MakeSetDisjointUnion(tm: TestManagerInterface) -> None:
    tm.addtest(check_set_disjoint_union)


@tfor(I.MakeSetDisjointUnion)
def check_set_disjoint_union(tc: TestContext) -> None:
    msd: I.MakeSetDisjointUnion = find_imp(tc, I.MakeSetDisjointUnion)
    set_empty = load_set_tc(tc, "set_empty")

    set2 = tc.check_result(msd, msd.disjoint_union, I.FiniteSetDisjointUnion, [set_empty, set_empty])
    check_same_set(tc, set2, set_empty)

    set_one: I.FiniteSet[int] = load_set_tc(tc, "set_one")
    set_two: I.FiniteSet[int] = load_set_tc(tc, "set_two")

    one_plus_two: I.FiniteSetDisjointUnion[int, Any] = tc.check_result(
        msd, msd.disjoint_union, I.FiniteSetDisjointUnion, [set_one, set_two]
    )
    elements = list(one_plus_two.elements())

    for e in elements:
        i, x = tc.check_result(one_plus_two, one_plus_two.unpack, object, e)
        e2 = tc.check_result(one_plus_two, one_plus_two.pack, object, i, x)
        check_same_element(tc, one_plus_two, e, e2)

    # either exception or in any case not allowing it
    try:
        banana = one_plus_two.pack(0, 42)
    except ValueError:
        pass
    else:
        tc.check_result_value(one_plus_two, one_plus_two.contains, bool, False, banana)

    with tc.description("Checking that bad indexes are not allowed."):
        # either exception or in any case not allowing it
        try:
            banana2 = one_plus_two.pack(-1, 42)
        except ValueError:
            pass
        else:
            tc.fail(zh.span("pack() should not allow negative index"), banana2=banana2)

        try:
            banana3 = one_plus_two.pack(3, 42)
        except ValueError:
            pass
        else:
            tc.fail(zh.span("pack() should not allow index too high"), banana3=banana3)

    with tc.description("Making sure it works for n = 0 sets."):
        zero = tc.check_result(msd, msd.disjoint_union, I.FiniteSetDisjointUnion, [])
        elements = list(zero.elements())
        tc.fail_not_equal2(0, len(elements), zh.span("Expected 0 elements"))



from abc import ABC, abstractmethod
from typing import Any, Generic, overload, Sequence, Tuple, TypeVar

from .sets import FiniteSet, Setoid

__all__ = ["SetDisjointUnion", "FiniteSetDisjointUnion", "MakeSetDisjointUnion"]

C = TypeVar("C")
E = TypeVar("E")


class SetDisjointUnion(Generic[C, E], Setoid[E], ABC):
    @abstractmethod
    def components(self) -> Sequence[Setoid[C]]:
        """Returns the components of the union"""

    @abstractmethod
    def pack(self, i: int, e: C) -> E:
        """Injection mapping: construct element of the i-th component.
        Raises InvalidValue if i is not a valid index or e is not an element of the i-th component.
        """

    @abstractmethod
    def unpack(self, e: E) -> Tuple[int, C]:
        """Returns the index of the component and the element."""


class FiniteSetDisjointUnion(Generic[C, E], FiniteSet[E], SetDisjointUnion[C, E], ABC):
    """Specialization of SetDisjointUnion where we deal with FiniteSets."""

    @abstractmethod
    def components(self) -> Sequence[FiniteSet[C]]:
        ...


class MakeSetDisjointUnion(ABC):
    @overload
    def disjoint_union(self, components: Sequence[FiniteSet[C]]) -> FiniteSetDisjointUnion[C, Any]:
        ...

    @overload
    def disjoint_union(self, components: Sequence[Setoid[C]]) -> SetDisjointUnion[C, Any]:
        ...

    @abstractmethod
    def disjoint_union(self, components: Sequence[Setoid[C]]) -> SetDisjointUnion[C, Any]:
        """Returns a disjoint union of the given components.
        Returns a finite set if all components are finite."""



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, List, TypeVar, Union

from typing_extensions import TypedDict

from .helper import IOHelper
from .sets import FiniteSet
from .types import ConcreteRepr

__all__ = [
    "FiniteSet_desc",
    "FiniteSetUnion_desc",
    "FiniteSetProduct_desc",
    "FiniteSetDisjointUnion_desc",
    "DirectElements_desc",
    "IOHelper",
    "FiniteSetRepresentation",
]

E = TypeVar("E")
E1 = TypeVar("E1")
E2 = TypeVar("E2")


class DirectElements_desc(TypedDict):
    elements: List[ConcreteRepr]


class FiniteSetProduct_desc(TypedDict):
    # product: List[FiniteSet_desc] # cannot be recursive
    product: List[Any]


class FiniteSetDisjointUnion_desc(TypedDict):
    # disunion: List[FiniteSet_desc]  # cannot be recursive
    disunion: List[Any]


class FiniteSetUnion_desc(TypedDict):
    # union: List[FiniteSet_desc]  # cannot be recursive
    union: List[Any]


FiniteSet_desc = Union[
    DirectElements_desc,
    FiniteSetProduct_desc,
    FiniteSetUnion_desc,
    FiniteSetDisjointUnion_desc,
]


class FiniteSetRepresentation(ABC):
    @abstractmethod
    def load(self, h: IOHelper, data: FiniteSet_desc) -> FiniteSet[Any]:
        """Load a finite set from data structure.
        Throw InvalidFormat if the format is incorrect.
        """

    @abstractmethod
    def save(self, h: IOHelper, f: FiniteSet[Any]) -> FiniteSet_desc:
        """Serializes into a data structure"""



from typing import TypedDict
from abc import ABC, abstractmethod
from .helper import IOHelper
from .profunctors import FiniteProfunctor


class FiniteProfunctor_desc(TypedDict):
    pass


class FiniteProfunctorRepresentation(ABC):
    @abstractmethod
    def load(self, h: IOHelper, data: FiniteProfunctor_desc) -> FiniteProfunctor:
        ...

    @abstractmethod
    def save(self, h: IOHelper, f: FiniteProfunctor) -> FiniteProfunctor_desc:
        ...



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, List

from typing_extensions import TypedDict

from .helper import IOHelper
from .relations import FiniteRelation
from .sets_representation import FiniteSet_desc
from .types import ConcreteRepr

__all__ = [
    "FiniteRelation_desc",
    "FiniteRelationRepresentation",
]


class FiniteRelation_desc(TypedDict):
    source: FiniteSet_desc
    target: FiniteSet_desc
    values: List[List[ConcreteRepr]]


class FiniteRelationRepresentation(ABC):
    @abstractmethod
    def load(self, h: IOHelper, data: FiniteRelation_desc) -> FiniteRelation[Any, Any]:
        """Load a finite set from given YAML data"""

    @abstractmethod
    def save(self, h: IOHelper, f: FiniteRelation[Any, Any]) -> FiniteRelation_desc:
        """Load a finite set from given YAML data"""



__all__ = [
    "InvalidFormat",
    "InvalidValue",
]


class InvalidFormat(Exception):
    """Raise this if the input data to parse is invalid."""


class InvalidValue(ValueError):
    """Raise this if the input does not make sense."""



from abc import ABC, abstractmethod
from typing import Any, Generic, Sequence, TypeVar

from .posets import FinitePoset, Poset
from .sets_product import (
    FiniteSetProduct,
    SetProduct,
)

__all__ = [
    "FinitePosetConstructionProduct",
    "FinitePosetProduct",
    "PosetProduct",
]

C = TypeVar("C")
E = TypeVar("E")


class PosetProduct(Generic[C, E], Poset[E], ABC):
    """A poset product is a poset that can be factorized."""

    @abstractmethod
    def carrier(self) -> SetProduct[C, E]:
        """Returns the components of the product"""

    @abstractmethod
    def components(self) -> Sequence[Poset[C]]:
        """Returns the components of the product"""


class FinitePosetProduct(Generic[C, E], FinitePoset[E], PosetProduct[C, E], ABC):
    """Specialization of PosetProduct where we deal with FinitePosets"""

    @abstractmethod
    def carrier(self) -> FiniteSetProduct[C, E]:
        """Returns the components of the product"""

    @abstractmethod
    def components(self) -> Sequence[FinitePoset[C]]:
        """Returns the components"""


class FinitePosetConstructionProduct(ABC):
    @abstractmethod
    def product(self, ps: Sequence[FinitePoset[C]]) -> FinitePosetProduct[C, Any]:
        ...



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from .categories import Category, FiniteCategory
from .functors import FiniteFunctor

__all__ = ["MonoidalCategory", "FiniteMonoidalCategory"]

Object = TypeVar("Object")
Morphism = TypeVar("Morphism")


class MonoidalCategory(Generic[Object, Morphism], Category[Object, Morphism], ABC):
    @abstractmethod
    def monoidal_unit(self) -> Object:
        """Return the product functor."""

    @abstractmethod
    def monoidal_product(self) -> FiniteFunctor[Object, Morphism, Object, Morphism]:  # XXX: to check
        """Return the product functor."""


class FiniteMonoidalCategory(
    Generic[Object, Morphism], MonoidalCategory[Object, Morphism], FiniteCategory[Object, Morphism], ABC
):
    ...



from typing import Any

import zuper_html as zh
from zuper_testint import find_imp, TestContext, tfor

import act4e_interfaces as I
from .posets_utils import check_same_poset, load_poset_tc


@tfor(I.FinitePosetConstructionSum)
def check_poset_disjoint_union(tc: TestContext) -> None:
    msd: I.FinitePosetConstructionSum = find_imp(tc, I.FinitePosetConstructionSum)
    poset_empty = load_poset_tc(tc, "poset_empty")

    poset2 = tc.check_result(msd, msd.disjoint_union, I.FinitePosetDisjointUnion, [poset_empty, poset_empty])
    # logger.info(poset2=poset2)

    check_same_poset(tc, poset2, poset_empty)

    set_one: I.FinitePoset[int] = load_poset_tc(tc, "poset_one")
    set_two: I.FinitePoset[int] = load_poset_tc(tc, "poset_two")

    one_plus_two: I.FinitePosetDisjointUnion[int, Any] = tc.check_result(
        msd, msd.disjoint_union, I.FinitePosetDisjointUnion, [set_one, set_two]
    )
    carrier = tc.check_result(one_plus_two, one_plus_two.carrier, I.FiniteSetDisjointUnion)
    elements = list(carrier.elements())

    for e in elements:
        one_plus_two.holds(e, e)

    with tc.description("Making sure it works for n = 0 sets."):
        zero = tc.check_result(msd, msd.disjoint_union, I.FinitePosetDisjointUnion, [])
        elements = list(zero.carrier().elements())
        tc.fail_not_equal2(0, len(elements), zh.span("Expected 0 elements"))



from abc import ABC, abstractmethod
from typing import Any, TypedDict, TypeVar

from .helper import IOHelper
from .maps import FiniteMap
from .maps_representation import FiniteMap_desc
from .semigroups import (
    FiniteGroup,
    FiniteGroupMorphism,
    FiniteMonoid,
    FiniteMonoidMorphism,
    FiniteSemigroup,
    FiniteSemigroupMorphism,
    GroupMorphism,
    MonoidMorphism,
    SemigroupMorphism,
)
from .semigroups_representation import FiniteGroup_desc, FiniteMonoid_desc, FiniteSemigroup_desc

A = TypeVar("A")
B = TypeVar("B")
C = TypeVar("C")


class FiniteSemigroupMorphismsChecks(ABC):
    @abstractmethod
    def is_semigroup_morphism(self, a: FiniteSemigroup[A], b: FiniteSemigroup[B], f: FiniteMap[A, B]) -> bool:
        """Check that the triple forms a semigroup morphism."""

    @abstractmethod
    def is_monoid_morphism(self, a: FiniteMonoid[A], b: FiniteMonoid[B], f: FiniteMap[A, B]) -> bool:
        """Check that the triple forms a monoid morphism."""

    @abstractmethod
    def is_group_morphism(self, a: FiniteGroup[A], b: FiniteGroup[B], f: FiniteMap[A, B]) -> bool:
        """Check that the triple forms a monoid morphism."""

    @abstractmethod
    def is_isomorphism(self, a: FiniteSemigroupMorphism[A, B]) -> bool:
        """Check that the morphism is an isomorphism."""


class FiniteSemigroupMorphismsOperations(ABC):
    @abstractmethod
    def compose_semi(self, f: SemigroupMorphism[A, B], g: SemigroupMorphism[B, C]) -> SemigroupMorphism[A, C]:
        ...

    @abstractmethod
    def compose_monoid(self, f: MonoidMorphism[A, B], g: MonoidMorphism[B, C]) -> MonoidMorphism[A, C]:
        ...

    @abstractmethod
    def compose_group(self, f: GroupMorphism[A, B], g: GroupMorphism[B, C]) -> GroupMorphism[A, C]:
        ...


class FiniteSemigroupMorphism_desc(TypedDict):
    source: FiniteSemigroup_desc
    target: FiniteSemigroup_desc
    mapping: FiniteMap_desc


class FiniteMonoidMorphism_desc(TypedDict):
    source: FiniteMonoid_desc
    target: FiniteMonoid_desc
    mapping: FiniteMap_desc


class FiniteGroupMorphism_desc(TypedDict):
    source: FiniteGroup_desc
    target: FiniteGroup_desc
    mapping: FiniteMap_desc


class FiniteSemigroupMorphismRepresentation(ABC):
    @abstractmethod
    def load(self, h: IOHelper, s: FiniteSemigroupMorphism_desc) -> FiniteSemigroupMorphism[Any, Any]:
        """Load the data"""

    @abstractmethod
    def save(self, h: IOHelper, m: FiniteSemigroupMorphism[Any, Any]) -> FiniteSemigroupMorphism_desc:
        """Save the data"""


class FiniteMonoidMorphismRepresentation(ABC):
    @abstractmethod
    def load(self, h: IOHelper, s: FiniteMonoidMorphism_desc) -> FiniteMonoidMorphism[Any, Any]:
        """Load the data"""

    @abstractmethod
    def save(self, h: IOHelper, m: FiniteMonoidMorphism[Any, Any]) -> FiniteMonoidMorphism_desc:
        """Save the data"""


class FiniteGroupMorphismRepresentation(ABC):
    @abstractmethod
    def load(self, h: IOHelper, s: FiniteGroupMorphism_desc) -> FiniteGroupMorphism[Any, Any]:
        """Load the data"""

    @abstractmethod
    def save(self, h: IOHelper, m: FiniteGroupMorphism[Any, Any]) -> FiniteGroupMorphism_desc:
        """Save the data"""



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Final, List, Union

from typing_extensions import TypedDict

from .helper import IOHelper
from .posets import FinitePoset
from .sets_representation import FiniteSet_desc
from .types import ConcreteRepr

__all__ = [
    "FinitePoset_desc",
    "FinitePoset_sum_desc",
    "FinitePoset_product_desc",
    "FinitePoset_simple_desc",
    "FinitePosetRepresentation",
]


class FinitePoset_simple_desc(TypedDict):
    carrier: FiniteSet_desc
    hasse: List[List[ConcreteRepr]]


class FinitePoset_sum_desc(TypedDict):
    poset_sum: List[Any]


class FinitePoset_product_desc(TypedDict):
    poset_product: List[Any]


FinitePoset_desc = Union[
    FinitePoset_simple_desc,
    FinitePoset_sum_desc,
    FinitePoset_product_desc,
]


class FinitePosetRepresentation(ABC):
    @abstractmethod
    def load(self, h: IOHelper, s: FinitePoset_desc) -> FinitePoset[Any]:
        """Load the data"""

    @abstractmethod
    def save(self, h: IOHelper, m: FinitePoset[Any]) -> FinitePoset_desc:
        """Save the data"""


KWD_POSET_SUM: Final[str] = "poset_sum"
KWD_POSET_PRODUCT: Final[str] = "poset_product"



from abc import ABC, abstractmethod
from typing import Any, Generic, overload, Sequence, TypeVar

from .posets import FinitePoset, Poset
from .sets_sum import (
    FiniteSetDisjointUnion,
    SetDisjointUnion,
)

__all__ = [
    "FinitePosetConstructionSum",
    "FinitePosetDisjointUnion",
    "PosetDisjointUnion",
]

E = TypeVar("E")
C = TypeVar("C")


class PosetDisjointUnion(Generic[C, E], Poset[E], ABC):
    @abstractmethod
    def carrier(self) -> SetDisjointUnion[C, E]:
        ...

    @abstractmethod
    def components(self) -> Sequence[Poset[C]]:
        ...


class FinitePosetDisjointUnion(Generic[C, E], FinitePoset[E], PosetDisjointUnion[C, E], ABC):
    """Specialization of PosetDisjointUnion where we deal with FiniteSets"""

    @abstractmethod
    def carrier(self) -> FiniteSetDisjointUnion[C, E]:
        ...

    @abstractmethod
    def components(self) -> Sequence[FinitePoset[C]]:
        ...


class FinitePosetConstructionSum(ABC):
    @overload
    def disjoint_union(self, ps: Sequence[FinitePoset[C]]) -> FinitePosetDisjointUnion[C, Any]:
        ...

    @overload
    def disjoint_union(self, ps: Sequence[Poset[C]]) -> PosetDisjointUnion[C, Any]:
        ...

    @abstractmethod
    def disjoint_union(self, ps: Sequence[Poset[C]]) -> PosetDisjointUnion[C, Any]:
        ...



from abc import ABC, abstractmethod
from typing import Any, TypedDict

from .adjunctions import FiniteAdjunction
from .helper import IOHelper

__all__ = ["FiniteAdjunction_desc", "FiniteAdjunctionRepresentation"]


class FiniteAdjunction_desc(TypedDict):
    pass


class FiniteAdjunctionRepresentation(ABC):
    @abstractmethod
    def load(self, h: IOHelper, data: FiniteAdjunction_desc) -> FiniteAdjunction[Any, Any, Any, Any]:
        ...

    @abstractmethod
    def save(self, h: IOHelper, f: FiniteAdjunction[Any, Any, Any, Any]) -> FiniteAdjunction_desc:
        ...



from functools import partial
from typing import Any, cast, List, TypeVar

import zuper_html as zh
from zuper_testint import ImplementationFail, TestContext, TestManagerInterface, TestRef, tfor

import act4e_interfaces as I
from .sets import check_save_all_elements
from .sets_utils import good_make_set

E = TypeVar("E")
X = TypeVar("X")


@tfor(I.MakeSetProduct)
def test_MakeSetProduct(tm: TestManagerInterface) -> None:
    A: I.FiniteSet[int] = good_make_set([1, 2])
    B: I.FiniteSet[str] = good_make_set(["a", "b"])

    AxB = tm_make_set_product(tm, A, B)

    tm.addtest(check_product, A, B, AxB)
    tm.addtest(check_save_all_elements, AxB)


def tm_make_set_product(tm: TestManagerInterface, *args: Any) -> TestRef[I.SetProduct[Any, Any]]:
    def f(tc: TestContext, imp: I.MakeSetProduct, *components: I.FiniteSet[Any]) -> I.SetProduct[Any, Any]:
        return tc.check_result(
            imp, imp.product, I.FiniteSetProduct, list(components)
        )  # return imp.product(list(components))

    i = tm.impof(I.MakeSetProduct)
    return tm.addtest(f, i, *args)


def check_product(
    tc: TestContext, A: I.FiniteSet[X], B: I.FiniteSet[X], AxB: I.FiniteSetProduct[Any, X]
) -> None:
    n = A.size() * B.size()
    tc.fail_not_equal2(n, AxB.size(), zh.span("Wrong size"))

    tc.f_expect_equal2(partial(AxB.contains, "Banana"), lambda: False, zh.span("Banana check"))

    a = list(A.elements())[0]
    b = list(B.elements())[0]

    ab = tc.check_result(AxB, AxB.pack, object, [a, b])
    unpacked = tc.check_result(AxB, AxB.unpack, List, ab)

    if len(unpacked) != 2:
        msg = zh.span("Expected unpacked to be a list of length 2")
        tc.fail(msg, unpacked=unpacked)
        raise ImplementationFail()
    tc.fail_not_equal2(unpacked[0], a, zh.span("Projection pA not working"))
    tc.fail_not_equal2(unpacked[1], b, zh.span("Projection pB not working"))

    # pA, pB = AxB.projections()
    #
    # tc.f_expect_equal2(L(pA, ab), lambda: a, zh.span("Projection pA not working"))
    # tc.f_expect_equal2(L(pB, ab), lambda: b, zh.span("Projection pB not working"))

    tc.f_expect_equal2(
        lambda: AxB.contains(cast(Any, 1)),
        lambda: False,
        zh.span('Expected that the element "1" is not contained'),
    )
    tc.f_expect_equal2(
        lambda: AxB.contains(ab), lambda: True, zh.span("Expected that the element is not contained")
    )



from abc import ABC, abstractmethod
from typing import Callable, Generic, Iterator, TypeVar

from .helper import IOHelper
from .types import ConcreteRepr

E = TypeVar("E")
E1 = TypeVar("E1")
E2 = TypeVar("E2")
E3 = TypeVar("E3")

__all__ = [
    "Setoid",
    "EnumerableSet",
    "FiniteSet",
    "EnumerableSetsOperations",
    "FiniteSetProperties",
    "EnumerableSetsOperations",
]


class Setoid(Generic[E], ABC):
    """
    A setoid is something to which elements may belong,
    has a way of distinguishing elements,
    and is able to (de)serialize its elements.
    """

    @abstractmethod
    def contains(self, x: E) -> bool:
        """Returns true if the element is in the set."""

    def equal(self, x: E, y: E) -> bool:
        """Returns True if the two elements are to be considered equal."""
        return x == y  # default is to use the Python equality

    def apart(self, x: E, y: E) -> bool:
        return not self.equal(x, y)

    @abstractmethod
    def save(self, h: IOHelper, x: E) -> ConcreteRepr:
        ...

    @abstractmethod
    def load(self, h: IOHelper, o: ConcreteRepr) -> E:
        ...


class EnumerableSet(Setoid[E], ABC):
    @abstractmethod
    def elements(self) -> Iterator[E]:
        """Note: possibly non-terminating."""


class FiniteSet(EnumerableSet[E], ABC):
    """A finite set has a finite size."""

    @abstractmethod
    def size(self) -> int:
        """Return the size of the finite set."""


class FiniteSetProperties(ABC):
    @abstractmethod
    def is_subset(self, a: FiniteSet[E], b: FiniteSet[E]) -> bool:
        """True if `a` is a subset of `b`."""

    def equal(self, a: FiniteSet[E], b: FiniteSet[E]) -> bool:
        return self.is_subset(a, b) and self.is_subset(b, a)

    def is_strict_subset(self, a: FiniteSet[E], b: FiniteSet[E]) -> bool:
        return self.is_subset(a, b) and not self.is_subset(b, a)


class EnumerableSetsOperations(ABC):
    @abstractmethod
    def make_set_sequence(self, f: Callable[[int], E]) -> EnumerableSet[E]:
        """Creates an EnumerableSet from a function that gives the
        i-th element."""

    @abstractmethod
    def union_esets(self, a: EnumerableSet[E], b: EnumerableSet[E]) -> EnumerableSet[E]:
        """Creates the union of two EnumerableSet."""



import itertools
from typing import Any, Callable, List, Sequence, TypeVar

import zuper_html as zh
from zuper_testint import find_imp, TestContext

import act4e_interfaces as I
from .actual_tests import good_fsr
from .data import get_test_sets, IOHelperImp, loadit_, purify_data

E = TypeVar("E")
X = TypeVar("X")


def check_lists_same(tc: TestContext, S: I.FiniteSet[X], s1: List[X], s2: List[X]) -> None:
    if len(s1) != len(s2):
        msg = "Lists have different number of elements"
        tc.fail(zh.span(msg), s1=s1, s2=s2)
        return

    for a in s1:
        included = any(tc.check_result(S, S.equal, bool, a, b) for b in s2)
        if not included:
            tc.fail(zh.span("Element not in the set"), a=a, s1=s1, s2=s2)


def load_list(tc: TestContext, S: I.FiniteSet[X], a: List[I.ConcreteRepr]) -> List[X]:
    h = IOHelperImp()
    s2 = [tc.check_result(S, S.load, object, h, _) for _ in a]
    return s2


def check_same_element(tc: TestContext, s1: I.FiniteSet[X], a: X, b: X) -> None:
    tc.check_result_value(
        s1, s1.equal, bool, True, a, b
    )  # # if not s1.equal(a, b):  #     msg = zh.span("Elements are not the same")  #     tc.fail(
    # msg, a=a, b=b)  #     raise ImplementationFail()


def check_same_set(tc: TestContext, s1: I.FiniteSet[X], s2: I.FiniteSet[X]) -> None:
    elements1 = list(s1.elements())
    elements2 = list(s2.elements())
    for e1 in elements1:
        belongs = tc.check_result(s2, s2.contains, bool, e1)
        if not belongs:
            msg = zh.span("Element in first set does not belong to the second set")
            tc.fail(msg, e1=e1)
    for e2 in elements2:
        belongs = tc.check_result(s1, s1.contains, bool, e2)
        if not belongs:
            msg = zh.span("Element in second set does not belong to the first set")
            tc.fail(msg, e2=e2)


def set_coherence(tc: TestContext, fi: I.FiniteSet[X]) -> None:
    a = len(list(fi.elements()))
    b = fi.size()
    if a != b:
        tc.fail(zh.p("size() and elements() are not coherent"))
    for element in fi.elements():
        if not fi.contains(element):
            tc.fail(zh.p("element in the set does not pass contains()"), element=element)
        if not fi.equal(element, element):
            tc.fail(zh.p("element in the set is not equal to itself"), element=element)


def make_values1(s: Sequence[X], f: Callable[[X], X]) -> List[List[X]]:
    return [[x, f(x)] for x in s]


def make_values(s: Sequence[X], f: Callable[[X, X], X]) -> List[Any]:
    return [[[x, y], f(x, y)] for x, y in itertools.product(s, s)]


def make_set(s: List[Any]) -> I.FiniteSet_desc:
    return {"elements": s}


def load_set(fsp: I.FiniteSetRepresentation, name: str) -> I.FiniteSet[Any]:
    d = get_test_sets()
    h = IOHelperImp()
    p1: I.FiniteSet[Any] = fsp.load(h, purify_data(d[name].data))
    return p1


def good_load_set(setname: str) -> I.FiniteSet[Any]:
    h = IOHelperImp()
    data = get_set_data(setname)

    return good_fsr.load(h, data)


def good_make_set(elements: Sequence[Any]) -> I.FiniteSet[Any]:
    h = IOHelperImp()
    return good_fsr.load(h, {"elements": list(elements)})


def load_set_tc(tc: TestContext, name: str) -> I.FiniteSet[Any]:
    fsr: I.FiniteSetRepresentation = find_imp(tc, I.FiniteSetRepresentation)
    data1 = get_set_data(name)
    h = IOHelperImp()
    return loadit_(tc, fsr, h, data1, I.FiniteSet)


def get_set_data(name: str) -> I.FiniteSet_desc:
    d = get_test_sets()
    data1 = purify_data(d[name].data)
    return data1



from typing import Any, cast

import act4e_interfaces as I
from .finite_set import MyFiniteSet

__all__ = ["MyFiniteSetRepresentation"]


class MyFiniteSetRepresentation(I.FiniteSetRepresentation):
    def load(self, h: I.IOHelper, data: I.FiniteSet_desc) -> I.FiniteSet[Any]:
        if not isinstance(data, dict):
            raise I.InvalidFormat(f"not a dictionary: {data}")
        if "elements" in data:
            data = cast(I.DirectElements_desc, data)
            elements = data["elements"]
            if not isinstance(elements, list):
                raise I.InvalidFormat()
            return MyFiniteSet(elements)
        raise I.InvalidFormat()

    def save(self, h: I.IOHelper, f: I.FiniteSet[Any]) -> I.FiniteSet_desc:
        elements = sorted(f.elements())
        data: I.FiniteSet_desc = {"elements": elements}
        return data



from abc import ABC
from dataclasses import dataclass
from typing import List

__all__ = ["Component", "Rope", "Elastic", "Brick", "Sphere", "Solution", "RupeGoldbergSolver"]


@dataclass
class Component:
    mass: float
    length: float
    spring_const_pull: float
    spring_const_push: float
    max_compression: float
    max_extension: float


@dataclass
class Rope(Component):
    length: float


@dataclass
class Elastic(Rope):
    elasticity: float  # meaning?


@dataclass
class Brick(Component):
    pass  # add other components / properties


@dataclass
class Sphere(Component):
    pass  # add other components / properties


@dataclass
class Solution:
    # does it break?
    breaks: bool
    # what is the total length (if not broken)?
    total_length: float


class RupeGoldbergSolver(ABC):
    def hangthem(self, components: List[Component]) -> Solution:
        """What if we hang the first component, and let the others hang below it?"""

    def push(self, components: List[Component], force: float) -> Solution:
        """What if we fix one endpoint, and we compress it?"""

    def pull(self, components: List[Component], force: float) -> Solution:
        """What if we fix one endpoint, and we pull the other end?"""



import os
from typing import List


def get_data_dir() -> str:
    from . import __file__

    d = os.path.join(os.path.dirname(__file__), "thedata")
    return d


def list_data_files() -> List[str]:
    d = get_data_dir()
    return [f for f in os.listdir(d) if f.endswith(".yaml")]


def load_test_data_file(fn: str) -> str:
    d = get_data_dir()

    fn2 = os.path.join(d, fn)
    if not os.path.exists(fn2):
        msg = f"Cannot find filename: {fn2}"
        raise ValueError(msg)

    with open(fn2) as f:
        return f.read()



__version__ = "7.3.2202231547"
__date__ = "2022-02-23T15:47:54+01:00"

from .adjunctions import *
from .adjunctions_representation import *
from .categories import *
from .categories_monoidal import *
from .categories_representation import *
from .dps import *
from .dps_representation import *
from .exceptions import *
from .functors import *
from .functors_representation import *
from .helper import *
from .intro import *
from .maps import *
from .maps_representation import *
from .natural import *
from .natural_representation import *
from .posets import *
from .posets_bounds import *
from .posets_construction import *
from .posets_interval import *
from .posets_lattices import *
from .posets_maps import *
from .posets_monoidal import *
from .posets_product import *
from .posets_representation import *
from .posets_sum import *
from .profunctors import *
from .profunctors_representation import *
from .relations import *
from .relations_representation import *

from .semigroups import *
from .semigroups_representation import *

from .sets import *
from .sets_power import *
from .sets_product import *
from .sets_representation import *
from .sets_sum import *
from .sets_union_inter import *
from .types import *
from .z_ex_rope import *
from .semigroups_operations import *



from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar

from .categories import Category, FiniteCategory
from .functors import FiniteFunctor, Functor
from .relations import FiniteRelation

__all__ = ["Adjunction", "FiniteAdjunction", "FiniteAdjunctionsOperations"]
Object = TypeVar("Object")
Morphism = TypeVar("Morphism")

Ob1 = TypeVar("Ob1")
Mor1 = TypeVar("Mor1")
Ob2 = TypeVar("Ob2")
Mor2 = TypeVar("Mor2")
Ob3 = TypeVar("Ob3")
Mor3 = TypeVar("Mor3")


class Adjunction(Generic[Ob1, Mor1, Ob2, Mor2], ABC):
    @abstractmethod
    def source(self) -> Category[Ob1, Mor1]:
        ...

    @abstractmethod
    def target(self) -> Category[Ob2, Mor2]:
        ...

    @abstractmethod
    def left(self) -> Functor[Ob1, Mor1, Ob2, Mor2]:
        ...

    @abstractmethod
    def right(self) -> Functor[Ob2, Mor2, Ob1, Mor1]:
        ...


class FiniteAdjunction(Generic[Ob1, Mor1, Ob2, Mor2], Adjunction[Ob1, Mor1, Ob2, Mor2], ABC):
    @abstractmethod
    def source(self) -> FiniteCategory[Ob1, Mor1]:
        ...

    @abstractmethod
    def target(self) -> FiniteCategory[Ob2, Mor2]:
        ...

    @abstractmethod
    def left(self) -> FiniteFunctor[Ob1, Mor1, Ob2, Mor2]:
        ...

    @abstractmethod
    def right(self) -> FiniteFunctor[Ob2, Mor2, Ob1, Mor1]:
        ...


class FiniteAdjunctionsOperations(ABC):
    @abstractmethod
    def is_adjunction(
        self,
        left: FiniteFunctor[Ob1, Mor1, Ob2, Mor2],
        right: FiniteFunctor[Ob2, Mor2, Ob1, Mor1],
    ) -> bool:
        """check the pair is an adjunction"""

    @abstractmethod
    def compose(
        self,
        adj1: FiniteAdjunction[Ob1, Mor1, Ob2, Mor2],
        adj2: FiniteAdjunction[Ob2, Mor2, Ob3, Mor3],
    ) -> FiniteAdjunction[Ob1, Mor1, Ob3, Mor3]:
        """compose two compatible adjunctions"""

    @abstractmethod
    def from_relation(
        self, f: FiniteRelation[Any, Any]
    ) -> FiniteAdjunction[Any, Any, Any, Any]:  # TODO: type
        ...



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from .posets import FinitePoset, Poset
from .posets_maps import MonotoneMap
from .relations import FiniteRelation
from .sets import FiniteSet, Setoid

__all__ = [
    "FiniteDP",
    "FiniteDPOperations",
    "DP",
    "DPI",
    "DPConstructors",
]


A = TypeVar("A")
B = TypeVar("B")
C = TypeVar("C")


F = TypeVar("F")
I = TypeVar("I")
R = TypeVar("R")


class DPI(Generic[F, I, R], ABC):
    @abstractmethod
    def functionality(self) -> Poset[F]:
        ...

    @abstractmethod
    def implementations(self) -> Setoid[I]:
        ...

    @abstractmethod
    def costs(self) -> Poset[R]:
        ...

    @abstractmethod
    def requires(self, i: I) -> R:
        ...

    @abstractmethod
    def provides(self, i: I) -> F:
        ...


class FiniteDPI(Generic[F, I, R], DPI[F, I, R], ABC):
    @abstractmethod
    def functionality(self) -> FinitePoset[F]:
        ...

    @abstractmethod
    def implementations(self) -> FiniteSet[I]:
        ...

    @abstractmethod
    def costs(self) -> FinitePoset[R]:
        ...


class DP(Generic[F, R], ABC):
    @abstractmethod
    def functionality(self) -> Poset[F]:
        ...

    @abstractmethod
    def costs(self) -> Poset[R]:
        ...

    @abstractmethod
    def feasible(self, f: F, r: R) -> bool:
        ...


class FiniteDP(Generic[F, R], DP[F, R], ABC):
    @abstractmethod
    def functionality(self) -> FinitePoset[F]:
        ...

    @abstractmethod
    def costs(self) -> FinitePoset[R]:
        ...


class DPConstructors(ABC):
    @abstractmethod
    def companion(self, f: MonotoneMap[A, B]) -> DP[A, B]:
        ...

    @abstractmethod
    def conjoint(self, f: MonotoneMap[A, B]) -> DP[A, B]:
        ...


class FiniteDPOperations(ABC):
    @abstractmethod
    def series(self, dp1: FiniteDP[A, B], dp2: FiniteDP[B, C]) -> FiniteDP[A, C]:
        ...

    @abstractmethod
    def union(self, dp1: FiniteDP[A, B], dp2: FiniteDP[A, B]) -> FiniteDP[A, B]:
        ...

    @abstractmethod
    def intersection(self, dp1: FiniteDP[A, B], dp2: FiniteDP[A, B]) -> FiniteDP[A, B]:
        ...

    @abstractmethod
    def from_relation(self, f: FiniteRelation[A, B]) -> FiniteDP[A, B]:
        ...



from abc import ABC, abstractmethod
from typing import Any, Generic, Iterator, overload, Sequence, TypeVar

from .sets import FiniteSet, Setoid

__all__ = ["MakePowerSet", "SetOfFiniteSubsets", "FiniteSetOfFiniteSubsets", "SetOfFiniteSubsets"]

C = TypeVar("C")
E = TypeVar("E")


class SetOfFiniteSubsets(Generic[C, E], Setoid[E], ABC):
    """A set of subsets."""

    @abstractmethod
    def contents(self, e: E) -> Iterator[C]:
        """Returns the contents of an element representing a subset."""

    @abstractmethod
    def construct(self, elements: Sequence[C]) -> E:
        """Get the element representing the given subset."""


class FiniteSetOfFiniteSubsets(Generic[C, E], SetOfFiniteSubsets[C, E], FiniteSet[E], ABC):
    pass


class MakePowerSet(ABC):
    @overload
    def powerset(self, s: FiniteSet[C]) -> FiniteSetOfFiniteSubsets[C, Any]:
        ...

    @overload
    def powerset(self, s: Setoid[C]) -> SetOfFiniteSubsets[C, Any]:
        ...

    @abstractmethod
    def powerset(self, s: Setoid[C]) -> SetOfFiniteSubsets[C, Any]:
        """Creates the powerset of a setoid. Returns a finite set if s is finite."""



from abc import ABC, abstractmethod
from typing import Any, List, TypedDict

from . import FiniteDP
from .posets_representation import FinitePoset_desc
from .sets_representation import FiniteSet_desc
from .types import ConcreteRepr

__all__ = [
    "FiniteDPRepresentation",
    "FiniteDP_desc",
]


class FiniteDP_desc(TypedDict):
    F: FinitePoset_desc
    R: FinitePoset_desc
    I: FiniteSet_desc
    feas: List[List[ConcreteRepr]]


class FiniteDPRepresentation(ABC):
    @abstractmethod
    def load(self, yaml_data: FiniteDP_desc) -> FiniteDP[Any, Any]:
        ...

    @abstractmethod
    def save(self, f: FiniteDP[Any, Any]) -> FiniteDP_desc:
        ...



from typing import (
    Any,
)

import zuper_html as zh
from zuper_testint import find_imp, TestContext, TestManagerInterface, tfor

import act4e_interfaces as I
from .posets_utils import load_poset_tc, poset_coherence


@tfor(I.FinitePosetConstructionArrow)
def test_FinitePosetConstructionArrow(tm: TestManagerInterface) -> None:
    _mks = tm.impof(I.FinitePosetConstructionArrow)
    for poset_name in ["poset_empty", "poset_one", "poset_two"]:
        tm.addtest(check_arrow1, poset_name, tid0=f"FinitePosetConstructionArrow-{poset_name}")


@tfor(I.FinitePosetConstructionTwisted)
def test_FinitePosetConstructionTwisted(tm: TestManagerInterface) -> None:
    _mks = tm.impof(I.FinitePosetConstructionTwisted)
    for poset_name in ["poset_empty", "poset_one", "poset_two"]:
        tm.addtest(check_twisted1, poset_name, tid0=f"FinitePosetConstructionTwisted-{poset_name}")


def check_arrow1(tc: TestContext, poset_name: str) -> None:
    mks: I.FinitePosetConstructionArrow = find_imp(tc, I.FinitePosetConstructionArrow)
    p = load_poset_tc(tc, poset_name)
    p_arrow = tc.check_result(mks, mks.arrow, I.FinitePosetOfIntervals, p)
    poset_coherence(tc, p_arrow)


@tfor(I.FinitePosetConstructionArrow)
def check_arrow_alphabet(tc: TestContext) -> None:
    mks: I.FinitePosetConstructionArrow = find_imp(tc, I.FinitePosetConstructionArrow)
    p = load_poset_tc(tc, "poset_alphabet")
    p_arrow: I.FinitePosetOfIntervals[str, Any] = tc.check_result(mks, mks.arrow, I.FinitePosetOfIntervals, p)
    poset_coherence(tc, p_arrow)

    # logger.info(elements=list(p_arrow.carrier().elements()))

    def construct(a: Any, b: Any) -> object:
        return tc.check_result(p_arrow, p_arrow.construct, object, a, b)

    def fail_construct(a: Any, b: Any) -> None:
        tc.check_result_raises(p_arrow, p_arrow.construct, ValueError, a, b)

    with tc.description("Should not allow to create bad interval [a,1]"):
        fail_construct("a", 1)

    with tc.description("Should not allow to create bad interval [1,a]"):
        fail_construct(1, "a")

    with tc.description("Should not allow to create unordered interval [b,a]"):
        fail_construct("b", "a")

    i1 = construct("a", "d")
    i2 = construct("c", "e")
    tc.check_result_value(p_arrow, p_arrow.holds, object, True, i1, i2)
    tc.check_result_value(p_arrow, p_arrow.holds, object, False, i2, i1)


@tfor(I.FinitePosetConstructionTwisted)
def check_twisted_alphabet(tc: TestContext) -> None:
    mks: I.FinitePosetConstructionTwisted = find_imp(tc, I.FinitePosetConstructionTwisted)
    p0 = load_poset_tc(tc, "poset_alphabet")
    P: I.FinitePosetOfIntervals[str, Any] = tc.check_result(mks, mks.twisted, I.FinitePosetOfIntervals, p0)
    poset_coherence(tc, P)

    # logger.info(elements=list(P.carrier().elements()))

    def construct(a: Any, b: Any) -> object:
        return tc.check_result(P, P.construct, object, a, b)

    def fail_construct(a: Any, b: Any) -> None:
        tc.check_result_raises(P, P.construct, ValueError, a, b)

    with tc.description("Should not allow to create bad interval [a,1]"):
        fail_construct("a", 1)

    with tc.description("Should not allow to create bad interval [1,a]"):
        fail_construct(1, "a")

    with tc.description("Should not allow to create unordered interval [b,a]"):
        fail_construct("b", "a")

    i1 = construct("e", "f")
    i2 = construct("a", "z")
    tc.check_result_value(P, P.holds, object, True, i1, i2)
    tc.check_result_value(P, P.holds, object, False, i2, i1)


def check_twisted1(tc: TestContext, poset_name: str) -> None:
    mks: I.FinitePosetConstructionArrow = find_imp(tc, I.FinitePosetConstructionArrow)
    p = load_poset_tc(tc, poset_name)
    p_twisted = tc.check_result(mks, mks.arrow, I.FinitePosetOfIntervals, p)
    poset_coherence(tc, p_twisted)


@tfor(I.FinitePosetConstructionArrow)
def test_twisted_one(tc: TestContext) -> None:
    mks: I.FinitePosetConstructionArrow = find_imp(tc, I.FinitePosetConstructionArrow)
    p = load_poset_tc(tc, "poset_one")
    p_twisted: I.FinitePosetOfIntervals[Any, Any] = tc.check_result(
        mks, mks.arrow, I.FinitePosetOfIntervals, p
    )
    pairs = [p_twisted.boundaries(_) for _ in p_twisted.carrier().elements()]

    tc.fail_not_equal2(pairs, [(1, 1)], zh.p("Expected only one element"))



from abc import ABC, abstractmethod
from typing import Any, Generic, overload, Tuple, TypeVar

from .posets import FinitePoset, Poset

__all__ = [
    "FinitePosetConstructionArrow",
    "FinitePosetConstructionTwisted",
    "PosetOfIntervals",
    "FinitePosetOfIntervals",
    "FinitePosetOfIntervals",
]

C = TypeVar("C")
E = TypeVar("E")


class PosetOfIntervals(Generic[C, E], Poset[E], ABC):
    """A poset of intervals."""

    @abstractmethod
    def construct(self, a: C, b: C) -> E:
        """
        Constructs an interval given the boundaries.

        Must throw ValueError if either a or b is not in the poset,
        or if (a <= b) does not hold.
        """

    @abstractmethod
    def boundaries(self, interval: E) -> Tuple[C, C]:
        """Returns the boundaries of an interval."""


class FinitePosetOfIntervals(Generic[C, E], PosetOfIntervals[C, E], FinitePoset[E], ABC):
    pass


class FinitePosetConstructionTwisted(ABC):
    @overload
    def twisted(self, s: FinitePoset[C]) -> FinitePosetOfIntervals[C, Any]:
        ...

    @overload
    def twisted(self, s: Poset[C]) -> PosetOfIntervals[C, Any]:
        ...

    @abstractmethod
    def twisted(self, s: Poset[C]) -> PosetOfIntervals[C, Any]:
        ...


class FinitePosetConstructionArrow(ABC):
    @overload
    def arrow(self, s: FinitePoset[C]) -> FinitePosetOfIntervals[C, Any]:
        ...

    @overload
    def arrow(self, s: Poset[C]) -> PosetOfIntervals[C, Any]:
        ...

    @abstractmethod
    def arrow(self, s: Poset[C]) -> PosetOfIntervals[C, Any]:
        ...



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, overload, TypeVar

from .sets import FiniteSet, Setoid

A = TypeVar("A")
B = TypeVar("B")
C = TypeVar("C")

__all__ = ["FiniteMapOperations", "FiniteMap", "Mapping"]


class Mapping(Generic[A, B], ABC):
    @abstractmethod
    def source(self) -> Setoid[A]:
        ...

    @abstractmethod
    def target(self) -> Setoid[B]:
        ...

    @abstractmethod
    def __call__(self, a: A) -> B:
        ...


class FiniteMap(Generic[A, B], Mapping[A, B], ABC):
    """A finite map is a mapping between two finite sets."""

    @abstractmethod
    def source(self) -> FiniteSet[A]:
        ...

    @abstractmethod
    def target(self) -> FiniteSet[B]:
        ...


class FiniteMapOperations(ABC):
    @overload
    def identity(self, s: FiniteSet[A]) -> FiniteMap[A, A]:
        ...

    @overload
    def identity(self, s: Setoid[A]) -> Mapping[A, A]:
        ...

    @abstractmethod
    def identity(self, s: Setoid[A]) -> Mapping[A, A]:
        """Returns the identity on a set."""

    @abstractmethod
    def compose(self, f: FiniteMap[A, B], g: FiniteMap[B, C]) -> FiniteMap[A, C]:
        """Compose two functions."""



import itertools
from typing import Any, List, TypeVar

import zuper_html as zh
from zuper_testint import find_imp, TestContext, TestManagerInterface, TestRef

import act4e_interfaces as I
from .data import dumpit_, get_test_data, get_test_posets, IOHelperImp, loadit_, purify_data
from .sets_utils import check_same_set

X = TypeVar("X")


def check_same_poset(tc: TestContext, s1: I.FinitePoset[X], s2: I.FinitePoset[X]) -> None:
    check_same_set(tc, s1.carrier(), s2.carrier())

    es = list(s1.carrier().elements())

    for a, b in itertools.product(es, es):
        h1 = tc.check_result(s1, s1.holds, bool, a, b)
        h2 = tc.check_result(s2, s2.holds, bool, a, b)
        if h1 != h2:
            tc.fail(zh.span("Different at"), a=a, b=b, h1=h1, h2=h2)

    # check_same_map(c, s1.holds(), s2.holds())


def check_poset_width(
    tc: TestContext, frp: I.FinitePosetMeasurement, s: I.FinitePoset[Any], width: int
) -> None:
    res = tc.check_result(frp, frp.width, int, s)
    if res != width:
        msg = zh.span(f"Expected width = {width}, obtained {res}.")
        tc.fail(msg)


def check_poset_height(
    tc: TestContext, frp: I.FinitePosetMeasurement, s: I.FinitePoset[Any], height: int
) -> None:
    res = tc.check_result(frp, frp.height, int, s)
    if res != height:
        msg = zh.span(f"Expected height = {height}, obtained {res}.")
        tc.fail(msg)


def check_is_upperset(
    tc: TestContext,
    frp: I.FinitePosetSubsetProperties2,
    s: I.FinitePoset[Any],
    subset: List[I.ConcreteRepr],
    expect: bool,
) -> None:
    # TODO: wrap for safety
    h = IOHelperImp()
    carrier = s.carrier()
    elements = [carrier.load(h, _) for _ in subset]
    res = tc.check_result(frp, frp.is_upper_set, bool, s, elements)

    if res != expect:
        msg = zh.span(f"Expected is_upper_set = {expect}, obtained {res}.")
        tc.fail(msg)


def check_is_lowerset(
    tc: TestContext,
    frp: I.FinitePosetSubsetProperties2,
    s: I.FinitePoset[Any],
    subset: List[I.ConcreteRepr],
    expect: bool,
) -> None:
    # TODO: wrap for safety
    h = IOHelperImp()
    carrier = s.carrier()
    elements = [carrier.load(h, _) for _ in subset]
    res = tc.check_result(frp, frp.is_lower_set, bool, s, elements)

    if res != expect:
        msg = zh.span(f"Expected is_lower_set = {expect}, obtained {res}.")
        tc.fail(msg)


def check_is_chain(
    tc: TestContext,
    frp: I.FinitePosetSubsetProperties,
    s: I.FinitePoset[Any],
    subset: List[I.ConcreteRepr],
    expect: bool,
) -> None:
    # TODO: wrap for safety
    h = IOHelperImp()
    carrier = s.carrier()
    elements = [carrier.load(h, _) for _ in subset]
    res = tc.check_result(frp, frp.is_chain, bool, s, elements)

    if res != expect:
        msg = zh.span(f"Expected is_antichain = {expect}, obtained {res}.")
        tc.fail(msg)


def check_is_antichain(
    tc: TestContext,
    frp: I.FinitePosetSubsetProperties,
    s: I.FinitePoset[Any],
    subset: List[I.ConcreteRepr],
    expect: bool,
) -> None:
    h = IOHelperImp()
    carrier = s.carrier()
    elements = [carrier.load(h, _) for _ in subset]
    res = tc.check_result(frp, frp.is_antichain, bool, s, elements)

    if res != expect:
        msg = zh.span(f"Expected is_antichain = {expect}, obtained {res}.")
        tc.fail(msg)


def load_the_poset(tm: TestManagerInterface, name: str) -> TestRef[I.FinitePoset[Any]]:
    d = get_test_posets()
    p1 = tm_load_poset(tm, name, tm.store(purify_data(d[name].data)))
    return p1


def tm_load_poset(
    tm: TestManagerInterface, name: str, data: TestRef[I.FinitePoset_desc]
) -> TestRef[I.FinitePoset[Any]]:
    h = IOHelperImp()

    def loadit(
        tc: TestContext, fsr: I.FinitePosetRepresentation, data1: I.FinitePoset_desc
    ) -> I.FinitePoset[Any]:
        return loadit_(tc, fsr, h, data1, I.FinitePoset)

    fsr_ = tm.impof(I.FinitePosetRepresentation)
    return tm.addtest(loadit, fsr_, data, tid0=f"load-poset-{name}")


def tm_save_poset(
    tm: TestManagerInterface, name: str, sgr: TestRef[I.FinitePoset[Any]]
) -> TestRef[I.FinitePoset_desc]:
    h = IOHelperImp()

    def dumpit(
        tc: TestContext, fsr: I.FinitePosetRepresentation, sgr_: I.FinitePoset[Any]
    ) -> I.FinitePoset_desc:
        return dumpit_(tc, fsr, h, sgr_)

    fsr_ = tm.impof(I.FinitePosetRepresentation)
    return tm.addtest(dumpit, fsr_, sgr, tid0=f"dump-poset-{name}")


def load_poset_tc(tc: TestContext, name: str) -> I.FinitePoset[Any]:
    fsr: I.FinitePosetRepresentation = find_imp(tc, I.FinitePosetRepresentation)
    data1 = get_poset_data(name)
    h = IOHelperImp()
    return loadit_(tc, fsr, h, data1, I.FinitePoset)


def get_poset_data(name: str) -> I.FinitePoset_desc:
    d = get_test_data("poset")
    data1 = purify_data(d[name].data)
    return data1


def poset_coherence(tc: TestContext, fi: I.FinitePoset[X]) -> None:
    carrier = tc.check_result(fi, fi.carrier, I.FiniteSet[X])
    elements = list(tc.check_result(fi, carrier.elements, object))
    for a in elements:
        tc.check_result_value(fi, fi.holds, bool, True, a, a)



from typing import (
    Any,
    Callable,
    cast,
    Dict,
    TypeVar,
)

import zuper_html as zh
from zuper_testint import (
    find_imp,
    ImplementationFail,
    TestContext,
    TestManagerInterface,
    TestNotImplemented,
    TestRef,
    tfor,
)

import act4e_interfaces as I
from .data import check_good_output, get_test_data, get_test_maps, IOHelperImp, loadit_, purify_data, TestData
from .relations import check_same_relation, load_relation_tc
from .sets_utils import check_same_element, check_same_set, load_set_tc, set_coherence

E = TypeVar("E")
A = TypeVar("A")
B = TypeVar("B")


@tfor(I.FiniteMapOperations)
def test_FiniteMapOperations_bool(tc: TestContext) -> None:
    fmo: I.FiniteMapOperations = find_imp(tc, I.FiniteMapOperations)

    with tc.description("Checking that we can create identities."):
        # Load the Bool set
        set_bool: I.FiniteSet[Any] = load_set_tc(tc, "set_bool")
        # Load the identity
        map_bool_identity_expected = load_map_tc(tc, "map_bool_identity")
        # Ask to create the identity
        map_bool_identity = cast(
            I.FiniteMap[Any, Any], tc.check_result(fmo, fmo.identity, I.FiniteMap, set_bool)
        )
        check_map(tc, map_bool_identity)
        check_same_map(tc, map_bool_identity, map_bool_identity_expected)

    with tc.description("Checking that we can compose maps"):
        map_bool_not = load_map_tc(tc, "map_bool_not")
        map_bool_not_twice = tc.check_result(fmo, fmo.compose, I.FiniteMap, map_bool_not, map_bool_not)
        check_map(tc, map_bool_not_twice)

        check_same_map(tc, map_bool_not_twice, map_bool_identity_expected)


@tfor(I.FiniteRelationOperations)
def test_FiniteRelationOperations_bool(tc: TestContext) -> None:
    fmo: I.FiniteRelationOperations = find_imp(tc, I.FiniteRelationOperations)
    with tc.description("Checking conversion of map to rel"):
        map_bool_not = load_map_tc(tc, "map_bool_not")
        rel_bool_not_expected = load_relation_tc(tc, "rel_bool_not")
        rel_bool_not = tc.check_result(fmo, fmo.as_relation, I.FiniteRelation, map_bool_not)
        check_same_relation(tc, rel_bool_not_expected, rel_bool_not)


@tfor(I.FiniteMapOperations)
def test_FiniteMapOperations(tm: TestManagerInterface) -> None:
    mks = tm.impof(I.FiniteMapOperations)

    raise TestNotImplemented()


def check_map(tc: TestContext, m: I.FiniteMap[A, B]) -> None:
    source = tc.check_result(m, m.source, I.FiniteSet)
    target = tc.check_result(m, m.target, I.FiniteSet)
    set_coherence(tc, source)
    set_coherence(tc, target)
    for a in source.elements():
        b = m.__call__(a)  # TODO: check it does not fail
        if not target.contains(b):
            tc.fail(zh.span("invalid result"), x=a, y=b)


def check_same_map(tc: TestContext, m1: I.FiniteMap[A, B], m2: I.FiniteMap[A, B]) -> None:
    m1_target = tc.check_call(m1.target, I.FiniteSet)
    m2_target = tc.check_call(m2.target, I.FiniteSet)
    m1_source = tc.check_call(m1.source, I.FiniteSet)
    m2_source = tc.check_call(m2.source, I.FiniteSet)

    check_same_set(tc, m1_source, m2_source)
    check_same_set(tc, m1_target, m2_target)
    for a in m1_source.elements():  # XXX: not wrapped
        # noinspection PyTypeChecker
        y1 = tc.check_call(m1, object, a)
        # noinspection PyTypeChecker
        y2 = tc.check_call(m2, object, a)

        equal = tc.check_call(m1_target.equal, bool, y1, y2)

        if not equal:
            tc.fail(zh.span("Maps have different results"), x=a, y1=y1, y2=y2)


def check_same_function(tc: TestContext, m1: I.FiniteMap[A, B], m2: I.FiniteMap[A, B]) -> None:
    check_same_set(tc, m1.source(), m2.source())
    check_same_set(tc, m1.target(), m2.target())
    S = m1.source()
    T = m1.target()
    for x1 in S.elements():
        y1 = m1.__call__(x1)
        y2 = m2.__call__(x1)
        check_same_element(tc, T, y1, y2)


@tfor(I.FiniteMapRepresentation)
def test_FiniteMapRepresentation(tm: TestManagerInterface) -> None:
    maps = get_test_data("map")

    doit_maps(tm, maps)


@tfor(I.FiniteMapRepresentation, level="Product")
def test_FiniteMapRepresentationProduct(tm: TestManagerInterface) -> None:
    d = get_test_data("map")
    res = {}
    for k, v in d.items():
        requires = dict(v.requires)
        requires.pop("set_product", None)
        if requires:
            continue
        res[k] = v
    doit_maps(tm, res)


def check_map_involutive(tc: TestContext, fs: I.FiniteSet[A], m: Callable[[A], A]) -> None:
    for e in fs.elements():
        e_inv = m(e)
        e2 = m(e_inv)
        equal = fs.equal(e, e2)
        if not equal:
            msg = zh.span("Not involutive")
            tc.fail(msg=msg, e=e, e_inv=e_inv, e2=e2)


def doit_maps(tm: TestManagerInterface, collection: Dict[str, TestData[I.FiniteMap_desc]]) -> None:
    def dump(
        tc: TestContext, fmr: I.FiniteMapRepresentation, fi: I.FiniteMap[A, B], keys: Dict[str, Any]
    ) -> I.FiniteMap_desc:
        h = IOHelperImp()
        try:
            res = fmr.save(h, fi)
        except BaseException as e:
            msg = "Cannot save the object"
            raise ImplementationFail(msg, fi=fi, **keys) from e
        check_good_output(tc, res)
        return res

    def load(
        tc: TestContext, fmr: I.FiniteMapRepresentation, data: I.FiniteMap_desc, keys: Dict[str, Any]
    ) -> I.FiniteMap[A, B]:
        h = IOHelperImp()
        try:
            return fmr.load(h, data)
        except I.InvalidFormat as e:
            msg = "The implementation could not read back the data that it wrote."
            raise ImplementationFail(msg, writtendata=data, **keys) from e

    _ = tm.impof(I.FiniteMapRepresentation)
    for mapname, md in collection.items():
        mapdata = tm.store(purify_data(md.data))
        m: TestRef[I.FiniteMap[Any, Any]] = tm_load_map(tm, mapname, mapdata)
        b = tm.addtest(dump, _, m, dict(mapname=mapname, orig_data=mapdata), tid0=f"map-{mapname}-dump")
        m2 = tm.addtest(load, _, b, dict(mapname=mapname, orig_data=mapdata), tid0=f"map-{mapname}-load2")

        tm.addtest(check_map, m)
        tm.addtest(check_map, m2)
        tm.addtest(check_same_map, m, m2)


def tm_load_map(
    tm: TestManagerInterface, name: str, data: TestRef[I.FiniteMap_desc]
) -> TestRef[I.FiniteMap[A, B]]:
    h = IOHelperImp()

    def loadit(tc: TestContext, fsr: I.FiniteMapRepresentation, data1: I.FiniteMap_desc) -> I.FiniteMap[A, B]:
        return loadit_(tc, fsr, h, data1, I.FiniteMap)

    _ = tm.impof(I.FiniteMapRepresentation)

    return tm.addtest(loadit, _, data, tid0=f"load-map-{name}")


def load_map_tc(tc: TestContext, name: str) -> I.FiniteMap[Any, Any]:
    fsr: I.FiniteMapRepresentation = find_imp(tc, I.FiniteMapRepresentation)
    data1 = get_map_data(name)
    h = IOHelperImp()
    return loadit_(tc, fsr, h, data1, I.FiniteMap)


def get_map_data(name: str) -> I.FiniteSet_desc:
    d = get_test_maps()
    data1 = purify_data(d[name].data)
    return data1



from abc import ABC, abstractmethod
from typing import Any, Generic, overload, Sequence, TypeVar

from .sets import EnumerableSet, FiniteSet, Setoid

__all__ = [
    "MakeSetIntersection",
    "MakeSetUnion",
    "SetUnion",
    "FiniteSetUnion",
    "EnumerableSetUnion",
    "SetoidOperations",
]

C = TypeVar("C")
E = TypeVar("E")


class MakeSetIntersection(ABC):
    @abstractmethod
    def intersection(self, components: Sequence[FiniteSet[E]]) -> FiniteSet[E]:
        ...


class SetUnion(Generic[C, E], Setoid[E], ABC):
    @abstractmethod
    def components(self) -> Sequence[Setoid[C]]:
        """Returns the components of the union"""


class EnumerableSetUnion(Generic[C, E], EnumerableSet[E], SetUnion[C, E], ABC):
    """Specialization of SetUnion where we deal with enumerable sets."""

    @abstractmethod
    def components(self) -> Sequence[EnumerableSet[C]]:
        """Returns the components of the union"""


class FiniteSetUnion(Generic[C, E], FiniteSet[E], EnumerableSetUnion[C, E], ABC):
    """Specialization of SetUnion where we deal with FiniteSets"""

    @abstractmethod
    def components(self) -> Sequence[FiniteSet[C]]:
        """Returns the components of the union"""


class MakeSetUnion(ABC):
    @overload
    def union(self, components: Sequence[FiniteSet[C]]) -> FiniteSetUnion[C, Any]:
        ...

    @overload
    def union(self, components: Sequence[EnumerableSet[C]]) -> EnumerableSetUnion[C, Any]:
        ...

    @abstractmethod
    def union(self, components: Sequence[Setoid[C]]) -> SetUnion[C, Any]:
        ...


class SetoidOperations(ABC):
    @abstractmethod
    def union_setoids(self, a: Setoid[E], b: Setoid[E]) -> Setoid[E]:
        """Creates the union of two Setoids."""

    @abstractmethod
    def intersection_setoids(self, a: Setoid[E], b: Setoid[E]) -> Setoid[E]:
        """Creates the intersection of two Setoids."""



from zuper_testint import find_imp, TestContext, tfor

import act4e_interfaces as I
from .maps import load_map_tc
from .posets_utils import load_poset_tc


@tfor(I.FiniteMonotoneMapProperties)
def check_FiniteMonotoneMapProperties(tc: TestContext) -> None:
    frc: I.FiniteMonotoneMapProperties = find_imp(tc, I.FiniteMonotoneMapProperties)

    with tc.description("Checking that we can compose relations"):
        map_bool_not = load_map_tc(tc, "map_bool_not")
        map_bool_identity = load_map_tc(tc, "map_bool_identity")
        poset_bool = load_poset_tc(tc, "poset_bool")

        with tc.description("not() is antitone"):
            tc.check_result_value(frc, frc.is_monotone, bool, False, poset_bool, poset_bool, map_bool_not)
            tc.check_result_value(frc, frc.is_antitone, bool, True, poset_bool, poset_bool, map_bool_not)

        with tc.description("identity() is monotone"):
            tc.check_result_value(frc, frc.is_monotone, bool, True, poset_bool, poset_bool, map_bool_identity)
            tc.check_result_value(
                frc, frc.is_antitone, bool, False, poset_bool, poset_bool, map_bool_identity
            )



from abc import ABC, abstractmethod
from typing import Any, Generic, overload, TypeVar

from .posets import FinitePoset, Poset
from .sets import FiniteSet, Setoid
from .sets_power import FiniteSetOfFiniteSubsets, SetOfFiniteSubsets

__all__ = [
    "FinitePosetConstructionOpposite",
    "PosetOfFiniteSubsets",
    "FinitePosetOfFiniteSubsets",
    "PosetConstructionPower",
]

C = TypeVar("C")
E = TypeVar("E")


class FinitePosetConstructionDiscrete(ABC):
    @overload
    def discrete(self, s: FiniteSet[E]) -> FinitePoset[E]:
        ...

    @overload
    def discrete(self, s: Setoid[E]) -> Poset[E]:
        ...

    @abstractmethod
    def discrete(self, s: Setoid[E]) -> Poset[E]:
        """Creates the discrete poset from any set."""


class PosetOfFiniteSubsets(Generic[C, E], Poset[E], ABC):
    """A poset of subsets."""

    @abstractmethod
    def carrier(self) -> SetOfFiniteSubsets[C, E]:
        ...


class FinitePosetOfFiniteSubsets(Generic[C, E], PosetOfFiniteSubsets[C, E], FinitePoset[E], ABC):
    @abstractmethod
    def carrier(self) -> FiniteSetOfFiniteSubsets[C, E]:
        ...


class PosetConstructionPower(ABC):
    @overload
    def powerposet(self, s: FiniteSet[C]) -> FinitePosetOfFiniteSubsets[C, Any]:
        ...

    @overload
    def powerposet(self, s: Setoid[C]) -> PosetOfFiniteSubsets[C, Any]:
        ...

    @abstractmethod
    def powerposet(self, s: Setoid[C]) -> PosetOfFiniteSubsets[C, Any]:
        ...


class FinitePosetConstructionOpposite(ABC):
    @overload
    def opposite(self, p: FinitePoset[E]) -> FinitePoset[E]:
        ...

    @overload
    def opposite(self, p: Poset[E]) -> Poset[E]:
        ...

    @abstractmethod
    def opposite(self, p: Poset[E]) -> Poset[E]:
        ...



from typing import (
    Any,
    List,
    Optional,
    TypeVar,
)

import zuper_html as zh
from zuper_testint import find_imp, TestContext, TestManagerInterface, tfor

import act4e_interfaces as I
from .data import get_test_posets, purify_data
from .posets_utils import (
    check_is_antichain,
    check_is_chain,
    check_is_lowerset,
    check_is_upperset,
    check_poset_height,
    check_poset_width,
    load_poset_tc,
    tm_load_poset,
)
from .sets_utils import check_lists_same, load_list


@tfor(I.FinitePosetMeasurement)
def test_FinitePosetMeasurements(tm: TestManagerInterface) -> None:
    mks = tm.impof(I.FinitePosetMeasurement)

    d = get_test_posets()
    for rname, rinfo in d.items():

        def get(p: Any) -> Any:
            return rinfo.properties.get(p, None)

        r1 = tm_load_poset(tm, rname, tm.store(purify_data(rinfo.data)))
        width = get("width")
        if width is not None:
            tm.addtest(check_poset_width, mks, r1, width, tid0=f"check_poset_width-{rname}")
        height = get("height")
        if height is not None:
            tm.addtest(check_poset_height, mks, r1, height, tid0=f"check_poset_height-{rname}")


@tfor(I.FinitePosetClosures)
def test_FinitePosetClosures(tm: TestManagerInterface) -> None:
    mks = tm.impof(I.FinitePosetClosures)

    d = get_test_posets()
    for rname, rinfo in d.items():

        def get(p: Any) -> Any:
            return rinfo.properties.get(p, None)

        s = get("some_upper_closures")
        if s is not None:
            for i, x in enumerate(s):
                tid = f"check_upper_closures-{rname}-{i}"
                tm.addtest(check_upper_closures, rname, x["S"], x["upper_closure"], tid0=tid)
        s = get("some_lower_closures")
        if s is not None:
            for i, x in enumerate(s):
                tid = f"check_upper_closures-{rname}-{i}"
                tm.addtest(check_lower_closures, rname, x["S"], x["lower_closure"], tid0=tid)


def check_lower_closures(
    tc: TestContext, poset_name: str, S_raw: List[I.ConcreteRepr], expected_raw: List[I.ConcreteRepr]
) -> None:
    P = load_poset_tc(tc, poset_name)
    carrier = tc.check_result(P, P.carrier, I.FiniteSet)
    S = load_list(tc, carrier, S_raw)
    expected = load_list(tc, carrier, expected_raw)

    fpc: I.FinitePosetClosures = find_imp(tc, I.FinitePosetClosures)
    cl = tc.check_result(fpc, fpc.lower_closure, list, P, S)
    with tc.description("checking correct lower closure", S=S, expected=expected, obtained=cl):
        check_lists_same(tc, carrier, expected, cl)


def check_upper_closures(
    tc: TestContext, poset_name: str, S_raw: List[I.ConcreteRepr], expected_raw: List[I.ConcreteRepr]
) -> None:
    P = load_poset_tc(tc, poset_name)
    carrier = tc.check_result(P, P.carrier, I.FiniteSet)
    S = load_list(tc, carrier, S_raw)
    expected = load_list(tc, carrier, expected_raw)

    fpc: I.FinitePosetClosures = find_imp(tc, I.FinitePosetClosures)
    cl = tc.check_result(fpc, fpc.upper_closure, list, P, S)
    with tc.description("checking correct upper closure", S=S, expected=expected, obtained=cl):
        check_lists_same(tc, carrier, expected, cl)


@tfor(I.FinitePosetSubsetProperties)
def test_FinitePosetSubsetProperties(tm: TestManagerInterface) -> None:
    mks = tm.impof(I.FinitePosetSubsetProperties)
    d = get_test_posets()
    for rname, rinfo in d.items():

        def get(p: Any) -> Any:
            return rinfo.properties.get(p, {})

        r1 = tm_load_poset(tm, rname, tm.store(purify_data(rinfo.data)))
        some_chains = get("some_chains")
        for k, v in some_chains.items():
            tm.addtest(check_is_chain, mks, r1, v, True, tid0=f"check_is_chain-{rname}-{k}")
        some_not_chains = get("some_not_chains")
        for k, v in some_not_chains.items():
            tm.addtest(check_is_chain, mks, r1, v, False, tid0=f"check_is_not_chain-{rname}-{k}")
        some_antichains = get("some_antichains")
        for k, v in some_antichains.items():
            tm.addtest(check_is_antichain, mks, r1, v, True, tid0=f"check_is_antichain-{rname}-{k}")
        some_not_antichains = get("some_not_antichains")
        for k, v in some_not_antichains.items():
            tm.addtest(check_is_antichain, mks, r1, v, False, tid0=f"check_is_not_antichain-{rname}-{k}")


@tfor(I.FinitePosetSubsetProperties2)
def test_FinitePosetSubsetProperties2(tm: TestManagerInterface) -> None:
    mks = tm.impof(I.FinitePosetSubsetProperties2)
    d = get_test_posets()
    for rname, rinfo in d.items():

        def get(p: Any) -> Any:
            return rinfo.properties.get(p, {})

        r1 = tm_load_poset(tm, rname, tm.store(purify_data(rinfo.data)))
        some_uppersets = get("some_uppersets")
        for k, v in some_uppersets.items():
            tm.addtest(check_is_upperset, mks, r1, v, True, tid0=f"check_is_upperset-{rname}-{k}")
        some_not_uppersets = get("some_not_uppersets")
        for k, v in some_not_uppersets.items():
            tm.addtest(check_is_upperset, mks, r1, v, False, tid0=f"check_is_not_upperset-{rname}-{k}")
        some_lowersets = get("some_lowersets")
        for k, v in some_lowersets.items():
            tm.addtest(check_is_lowerset, mks, r1, v, True, tid0=f"check_is_lowerset-{rname}-{k}")
        some_not_uppersets = get("some_not_lowersets")
        for k, v in some_not_uppersets.items():
            tm.addtest(check_is_lowerset, mks, r1, v, False, tid0=f"check_is_not_lowerset-{rname}-{k}")


@tfor(I.FinitePosetInfSup)
def test_FinitePosetInfSup(tm: TestManagerInterface) -> None:
    mks = tm.impof(I.FinitePosetInfSup)
    d = get_test_posets()
    for rname, rinfo in d.items():

        def get(p: Any) -> Any:
            return rinfo.properties.get(p, [])

        r1 = tm_load_poset(tm, rname, tm.store(purify_data(rinfo.data)))
        for i, v in enumerate(get("some_lower_bounds")):
            tm.addtest(
                check_lower_bounds, mks, r1, v["S"], v["lower_bounds"], tid0=f"check_lower_bounds-{rname}-{i}"
            )
        for i, v in enumerate(get("some_upper_bounds")):
            tm.addtest(
                check_upper_bounds, mks, r1, v["S"], v["upper_bounds"], tid0=f"check_upper_bounds-{rname}-{i}"
            )
        for i, v in enumerate(get("some_supremum")):
            tm.addtest(check_supremum, mks, r1, v["S"], v["supremum"], tid0=f"check_supremum-{rname}-{i}")
        for i, v in enumerate(get("some_infimum")):
            tm.addtest(check_infimum, mks, r1, v["S"], v["infimum"], tid0=f"check_infimum-{rname}-{i}")


@tfor(I.FinitePosetMinMax)
def test_FinitePosetMinMax(tm: TestManagerInterface) -> None:
    mks = tm.impof(I.FinitePosetMinMax)
    d = get_test_posets()
    for rname, rinfo in d.items():

        def get(p: Any) -> Any:
            return rinfo.properties.get(p, [])

        r1 = tm_load_poset(tm, rname, tm.store(purify_data(rinfo.data)))
        for i, v in enumerate(get("some_miminal")):
            tm.addtest(check_minimal, mks, r1, v["S"], v["minimal"], tid0=f"check_minimal-{rname}-{i}")
        for i, v in enumerate(get("some_maximal")):
            tm.addtest(check_maximal, mks, r1, v["S"], v["maximal"], tid0=f"check_maximal-{rname}-{i}")


X = TypeVar("X")


def check_infimum(
    tc: TestContext,
    mks: I.FinitePosetInfSup,
    p: I.FinitePoset[X],
    S_raw: List[I.ConcreteRepr],
    expected_raw: Optional[I.ConcreteRepr],
) -> None:
    carrier = p.carrier()
    S = load_list(tc, carrier, S_raw)
    result = tc.check_result(mks, mks.infimum, object, p, S)
    _check(tc, carrier, S, result, expected_raw)


def check_supremum(
    tc: TestContext,
    mks: I.FinitePosetInfSup,
    p: I.FinitePoset[X],
    S_raw: List[I.ConcreteRepr],
    expected_raw: Optional[I.ConcreteRepr],
) -> None:
    carrier = p.carrier()
    S = load_list(tc, carrier, S_raw)
    result = tc.check_result(mks, mks.supremum, object, p, S)
    _check(tc, carrier, S, result, expected_raw)


def _check(
    tc: TestContext, carrier: I.FiniteSet[X], S: List[X], result: X, expected_raw: Optional[I.ConcreteRepr]
) -> None:
    expected = load_list(tc, carrier, [expected_raw])[0] if expected_raw is not None else None
    if result is None and expected is not None:
        msg = "Expected existing but got None"
        tc.fail(zh.p(msg), S=S, expected=expected, result=result)
        return
    if result is not None and expected is None:
        msg = "Expected not existing but got some"
        tc.fail(zh.p(msg), S=S, expected=expected, result=result)
        return

    same = tc.check_result(carrier, carrier.equal, bool, result, expected)
    if not same:
        msg = "mismatch"
        tc.fail(zh.p(msg), S=S, expected=repr(expected), result=repr(result), same=same, carrier=carrier)


def check_maximal(
    tc: TestContext,
    mks: I.FinitePosetMinMax,
    p: I.FinitePoset[X],
    S_raw: List[I.ConcreteRepr],
    expected_raw: List[I.ConcreteRepr],
) -> None:
    carrier = p.carrier()
    S = load_list(tc, carrier, S_raw)
    result = tc.check_result(mks, mks.maximal, Optional[List[Any]], p, S)
    expected = load_list(tc, carrier, expected_raw)
    check_lists_same(tc, carrier, result, expected)


def check_minimal(
    tc: TestContext,
    mks: I.FinitePosetMinMax,
    p: I.FinitePoset[X],
    S_raw: List[I.ConcreteRepr],
    expected_raw: List[I.ConcreteRepr],
) -> None:
    carrier = p.carrier()
    S = load_list(tc, carrier, S_raw)
    result = tc.check_result(mks, mks.minimal, Optional[List[Any]], p, S)
    expected = load_list(tc, carrier, expected_raw)
    check_lists_same(tc, carrier, result, expected)


def check_lower_bounds(
    tc: TestContext,
    mks: I.FinitePosetInfSup,
    p: I.FinitePoset[X],
    S_raw: List[I.ConcreteRepr],
    expected_raw: List[I.ConcreteRepr],
) -> None:
    carrier = p.carrier()
    S = load_list(tc, carrier, S_raw)
    result = tc.check_result(mks, mks.lower_bounds, Optional[List[Any]], p, S)
    expected = load_list(tc, carrier, expected_raw)
    check_lists_same(tc, carrier, result, expected)


def check_upper_bounds(
    tc: TestContext,
    mks: I.FinitePosetInfSup,
    p: I.FinitePoset[X],
    S_raw: List[I.ConcreteRepr],
    expected_raw: List[I.ConcreteRepr],
) -> None:
    carrier = p.carrier()
    S = load_list(tc, carrier, S_raw)
    result = tc.check_result(mks, mks.upper_bounds, List[Any], p, S)
    expected = load_list(tc, carrier, expected_raw)
    check_lists_same(tc, carrier, result, expected)



from dataclasses import dataclass
from typing import cast, Generic, Iterator, List, TypeVar

import act4e_interfaces as I

__all__ = ["MyFiniteSet"]
X = TypeVar("X")


@dataclass
class MyFiniteSet(Generic[X], I.FiniteSet[X]):
    def save(self, h: I.IOHelper, x: X) -> I.ConcreteRepr:
        return cast(I.ConcreteRepr, x)

    def load(self, h: I.IOHelper, o: I.ConcreteRepr) -> X:
        return cast(X, o)

    _elements: List[X]

    def size(self) -> int:
        return len(self._elements)

    def contains(self, x: X) -> bool:
        for y in self._elements:
            if self.equal(x, y):
                return True
        return False

    def elements(self) -> Iterator[X]:
        for _ in self._elements:
            yield _



from zuper_testint import TestManagerInterface, TestNotImplemented, tfor

import act4e_interfaces as I


@tfor(I.FiniteNaturalTransformationRepresentation)
def test_FiniteNaturalTransformationRepresentation(tm: TestManagerInterface) -> None:
    mks = tm.impof(I.FiniteNaturalTransformationRepresentation)
    raise TestNotImplemented()


@tfor(I.FiniteFunctorRepresentation)
def test_FiniteFunctorRepresentation(tm: TestManagerInterface) -> None:
    mks = tm.impof(I.FiniteFunctorRepresentation)
    raise TestNotImplemented()


@tfor(I.FiniteProfunctorRepresentation)
def test_FiniteProfunctorRepresentation(tm: TestManagerInterface) -> None:
    mks = tm.impof(I.FiniteProfunctorRepresentation)
    raise TestNotImplemented()


@tfor(I.FiniteDPRepresentation)
def test_FiniteDPRepresentation(tm: TestManagerInterface) -> None:
    mks = tm.impof(I.FiniteDPRepresentation)

    raise TestNotImplemented()


@tfor(I.CategoryOperations)
def test_CategoryOperations(tm: TestManagerInterface) -> None:
    mks = tm.impof(I.CategoryOperations)
    raise TestNotImplemented()


@tfor(I.FiniteAdjunctionsOperations)
def test_FiniteAdjunctionsOperations(tm: TestManagerInterface) -> None:
    mks = tm.impof(I.FiniteAdjunctionsOperations)
    raise TestNotImplemented()


@tfor(I.FiniteDPOperations)
def test_FiniteDPOperations(tm: TestManagerInterface) -> None:
    mks = tm.impof(I.FiniteDPOperations)
    raise TestNotImplemented()


@tfor(I.FiniteProfunctorOperations)
def test_FiniteProfunctorOperations(tm: TestManagerInterface) -> None:
    mks = tm.impof(I.FiniteProfunctorOperations)
    raise TestNotImplemented()


@tfor(I.FiniteAdjunctionRepresentation)
def test_FiniteAdjunctionRepresentation(tm: TestManagerInterface) -> None:
    mks = tm.impof(I.FiniteAdjunctionRepresentation)
    raise TestNotImplemented()



import itertools
from functools import partial
from typing import (
    Any,
    Callable,
    List,
    TypeVar,
)

import zuper_html as zh
from zuper_commons.types import check_isinstance
from zuper_testint import find_imp, ImplementationFail, TestContext, TestManagerInterface, TestRef, tfor

import act4e_interfaces as I
from .data import dumpit_, get_test_data, IOHelperImp, loadit_, purify_data
from .maps import check_map_involutive
from .sets_utils import check_same_set, good_load_set, make_set, make_values

E = TypeVar("E")
X = TypeVar("X")


@tfor(I.FiniteSemigroupConstruct)
def test_FiniteSemigroupConstruct(tm: TestManagerInterface) -> None:
    mks = tm.impof(I.FiniteSemigroupConstruct)

    A = good_load_set("set1")
    B = good_load_set("set2")
    for s in (A, B):
        tm.addtest(check_free_semigroup, mks, s)


def check_free_semigroup(tc: TestContext, fsg: I.FiniteSemigroupConstruct, s: I.FiniteSet[X]) -> None:
    fs = tc.check_result(fsg, fsg.free, I.FreeSemigroup, s)
    carrier = fs.carrier()

    for a0, b0 in itertools.product(s.elements(), s.elements()):
        a = tc.check_result(fs, fs.unit, object, a0)
        b = tc.check_result(fs, fs.unit, object, b0)
        res = tc.check_result(fs, fs.compose, object, a, b)
        if not carrier.contains(res):
            msg = "Composition result does not belong to set."
            raise ImplementationFail(msg, a=a, b=b, res=res)


def check_semigroup(tc: TestContext, m: I.FiniteSemigroup[E]) -> None:
    carrier = m.carrier()
    elements = list(carrier.elements())
    for e1, e2 in itertools.product(elements, elements):
        e = tc.check_result(m, m.compose, object, e1, e2)
        inside = tc.check_result(carrier, carrier.contains, object, e)
        if not inside:
            tc.fail(zh.span("composition fails"), e1=e1, e2=e2, e=e, inside=inside)


def check_monoid(tc: TestContext, m: I.FiniteMonoid[E]) -> None:
    check_semigroup(tc, m)
    m.identity()


def check_group(tc: TestContext, m: I.FiniteGroup[E]) -> None:
    with tc.description("checking it is a valid monoid"):
        check_monoid(tc, m)

    with tc.description("checking group composition"):
        check_group_composition(tc, m)

    with tc.description("checking that the inverse is involutive"):
        check_map_involutive(tc, m.carrier(), m.inverse)


def check_group_composition(tc: TestContext, G: I.FiniteGroup[E]) -> None:
    carrier = G.carrier()
    neutral = G.identity()
    for e in carrier.elements():
        e_inv = G.inverse(e)  # TODO: wrap
        e_ei = G.compose(e, e_inv)
        e_ie = G.compose(e_inv, e)

        with tc.description("Checking that id = e*inv(e)"):
            tc.f_expect_equal2(
                partial(carrier.equal, neutral, e_ei),
                lambda: True,
                zh.span("id != e*i"),
                neutral=neutral,
                e=e,
                e_inv=e_inv,
                e_times_e_inv=e_ei,
            )
        with tc.description("Checking that id = inv(e)*e"):
            tc.f_expect_equal2(
                partial(carrier.equal, neutral, e_ie),
                lambda: True,
                zh.span("id != i*e"),
                neutral=neutral,
                e=e,
                e_inv=e_inv,
                e_inv_times_e=e_ie,
            )


def check_same_semigroup(tc: TestContext, g1: I.FiniteSemigroup[X], g2: I.FiniteSemigroup[X]) -> None:
    carrier = g1.carrier()
    check_same_set(tc, g1.carrier(), g2.carrier())
    elements = list(carrier.elements())

    for a, b in itertools.product(elements, elements):
        r1 = g1.compose(a, b)
        r2 = g2.compose(a, b)
        eq = carrier.equal(r1, r2)
        if not eq:
            msg = zh.span("Not equal at ")
            tc.fail(msg, a=a, b=b, r1=r1, r2=r2)


def check_same_monoid(tc: TestContext, g1: I.FiniteMonoid[X], g2: I.FiniteMonoid[X]) -> None:
    check_same_semigroup(tc, g1, g2)
    id1 = g1.identity()
    id2 = g2.identity()
    carrier = g1.carrier()
    eq = carrier.equal(id1, id2)
    if not eq:
        msg = zh.span("Different identity")
        tc.fail(msg, id1=id1, id2=id2)


def check_same_group(tc: TestContext, g1: I.FiniteGroup[X], g2: I.FiniteGroup[X]) -> None:
    check_same_monoid(tc, g1, g2)
    carrier = g1.carrier()

    for a in carrier.elements():
        r1 = g1.inverse(a)
        r2 = g2.inverse(a)
        eq = carrier.equal(r1, r2)
        if not eq:
            msg = zh.span("Inverse not equal to  a")
            tc.fail(msg, a=a, r1=r1, r2=r2)


def tm_load_semigroup(
    tm: TestManagerInterface,
    name: str,
    data: TestRef[I.FiniteSemigroup_desc],
) -> TestRef[I.FiniteSemigroup[Any]]:
    h = IOHelperImp()

    # if 'data' in data:
    #     raise ValueError(data)

    def loadit(
        c: TestContext, fsr: I.FiniteSemigroupRepresentation, data1: I.FiniteSemigroup_desc
    ) -> I.FiniteSemigroup[Any]:
        return loadit_(c, fsr, h, data1, I.FiniteSemigroup)

    fsr_ = tm.impof(I.FiniteSemigroupRepresentation)
    return tm.addtest(loadit, fsr_, data, tid0=f"load-semigroup-{name}")


def tm_save_semigroup(
    tm: TestManagerInterface, name: str, sgr: TestRef[I.FiniteSemigroup[Any]]
) -> TestRef[I.FiniteSemigroup_desc]:
    h = IOHelperImp()

    def dumpit(
        c: TestContext, fsr: I.FiniteSemigroupRepresentation, sgr_: I.FiniteSemigroup[Any]
    ) -> I.FiniteSemigroup_desc:
        return dumpit_(c, fsr, h, sgr_)

    fsr_ = tm.impof(I.FiniteSemigroupRepresentation)
    return tm.addtest(dumpit, fsr_, sgr, tid0=f"save-semigroup-{name}")


def tm_save_monoid(
    tm: TestManagerInterface, name: str, sgr: TestRef[I.FiniteMonoid[Any]]
) -> TestRef[I.FiniteMonoid_desc]:
    h = IOHelperImp()

    def dumpit(
        c: TestContext, fsr: I.FiniteMonoidRepresentation, sgr_: I.FiniteMonoid[Any]
    ) -> I.FiniteMonoid_desc:
        return dumpit_(c, fsr, h, sgr_)

    fsr_ = tm.impof(I.FiniteMonoidRepresentation)
    return tm.addtest(dumpit, fsr_, sgr, tid0=f"save-monoid-{name}")


def tm_save_group(
    tm: TestManagerInterface, name: str, sgr: TestRef[I.FiniteGroup[Any]]
) -> TestRef[I.FiniteGroup_desc]:
    h = IOHelperImp()

    def dumpit(
        tc: TestContext, fsr: I.FiniteGroupRepresentation, sgr_: I.FiniteGroup[Any]
    ) -> I.FiniteGroup_desc:
        return dumpit_(tc, fsr, h, sgr_)

    fsr_ = tm.impof(I.FiniteGroupRepresentation)
    return tm.addtest(dumpit, fsr_, sgr, tid0=f"dump-group-{name}")


def tm_load_group(
    tm: TestManagerInterface, name: str, data: TestRef[I.FiniteGroup_desc]
) -> TestRef[I.FiniteGroup[Any]]:
    h = IOHelperImp()

    def loadit(
        tc: TestContext, fsr: I.FiniteGroupRepresentation, data1: I.FiniteGroup_desc
    ) -> I.FiniteGroup[Any]:
        return loadit_(tc, fsr, h, data1, I.FiniteGroup)

    fsr_ = tm.impof(I.FiniteGroupRepresentation)
    return tm.addtest(loadit, fsr_, data, tid0=f"load-group-{name}")


def tm_load_monoid(
    tm: TestManagerInterface, name: str, data: TestRef[I.FiniteMonoid_desc]
) -> TestRef[I.FiniteMonoid[Any]]:
    h = IOHelperImp()

    def loadit(
        c: TestContext, fsr: I.FiniteMonoidRepresentation, data1: I.FiniteMonoid_desc
    ) -> I.FiniteMonoid[Any]:
        return loadit_(c, fsr, h, data1, I.FiniteMonoid)

    fsr_ = tm.impof(I.FiniteMonoidRepresentation)
    return tm.addtest(loadit, fsr_, data, tid0=f"load-monoid-{name}")


def make_semigroup(s: List[X], f: Callable[[X, X], X]) -> I.FiniteSemigroup_desc:
    check_isinstance(s, list)
    fs = make_set(s)
    res: I.FiniteSemigroup_desc = {
        "carrier": fs,
        "composition": make_values(s, f),
    }
    # logger.info(res=res)
    return res


@tfor(I.FiniteMonoidRepresentation)
def test_FiniteMonoidRepresentation(tm: TestManagerInterface) -> None:
    monoids = get_test_data("monoid")

    for name, m in monoids.items():
        data = m.data
        mon = tm_load_monoid(tm, name, data)
        mon2 = tm_save_monoid(tm, name, mon)
        tm_load_monoid(tm, name + "-re", mon2)
        tm.addtest(check_monoid, mon)


@tfor(I.FiniteSemigroupRepresentation)
def test_FiniteSemigroupRepresentation(tm: TestManagerInterface) -> None:
    semigroups = get_test_data("semigroup")
    for sgrp_name, s in semigroups.items():
        data = s.data

        fs = tm_load_semigroup(tm, sgrp_name, data)
        data2 = tm_save_semigroup(tm, sgrp_name, fs)
        fs = tm_load_semigroup(tm, sgrp_name + "-re", data2)


@tfor(I.FiniteGroupRepresentation)
def test_FiniteGroupRepresentation(tm: TestManagerInterface) -> None:
    groups = get_test_data("group")

    for name in groups:
        tm.addtest(check_group_load_save, name, tid0=f"check_group_load_save-{name}")


def check_group_load_save(tc: TestContext, name: str) -> None:
    desc = f"Checking if we can save/load the test group called {name}"
    with tc.description(desc):
        with tc.description(f"Trying to load the group {name}"):
            g1 = load_group_tc(tc, name)

        with tc.description("Checking it is a valid group"):
            check_group(tc, g1)

        with tc.description("Trying to save the group"):
            g1_dumped = save_group_tc(tc, g1)

        with tc.description("Trying to re-load the saved group"):
            fgr: I.FiniteGroupRepresentation = find_imp(tc, I.FiniteGroupRepresentation)
            h = IOHelperImp()
            g2 = fgr.load(h, g1_dumped)

        with tc.description("Checking the reloaded is a valid group"):
            check_group(tc, g1)

        tc.raise_if_failures()
        with tc.description("Checking the original and reloaded groups are the same"):
            check_same_group(tc, g1, g2)


def save_group_tc(tc: TestContext, g: I.FiniteGroup[Any]) -> I.FiniteGroup_desc:
    fgr: I.FiniteGroupRepresentation = find_imp(tc, I.FiniteGroupRepresentation)
    h = IOHelperImp()

    return dumpit_(tc, fgr, h, g)


def get_group_data(name: str) -> I.ConcreteRepr:
    d = get_test_data("group")
    data1 = purify_data(d[name].data)
    return data1


def load_group_tc(tc: TestContext, name: str) -> I.FiniteGroup[Any]:
    fgr: I.FiniteGroupRepresentation = find_imp(tc, I.FiniteGroupRepresentation)
    h = IOHelperImp()
    data1 = get_group_data(name)
    # not sure why
    return loadit_(tc, fgr, h, data1, I.FiniteGroup)  # type: ignore



import traceback
from dataclasses import dataclass
from datetime import datetime
from functools import lru_cache
from io import StringIO
from typing import Any, Dict, Generic, List, Protocol, TypeVar

import ruamel.yaml as yaml
import zuper_html as zh
from ruamel.yaml import RoundTripLoader, YAML
from zuper_commons.types import add_context, ZValueError
from zuper_testint import ImplementationFail, TestContext

import act4e_interfaces as I
from . import logger

X = TypeVar("X")


@dataclass
class TestData(Generic[X]):
    tags: Dict[str, bool]
    requires: Dict[str, bool]
    data: X
    properties: Dict[str, Any]


ALLOWED_TAGS = {
    "poset",
    "set",
    "semigroup",
    "monoid",
    "group",
    "relation",
    "map",
    "dp",
    "category",
    "natural_transform",
}
ALLOWED_PROPERTIES = {
    "powerset",
    "some_antichains",
    "some_not_antichains",
    "some_chains",
    "some_not_chains",
    "surjective",
    "defined_everywhere",
    "single_valued",
    "injective",
    "reflexive",
    "irreflexive",
    "transitive",
    "asymmetric",
    "symmetric",
    "antisymmetric",
    "has_top",
    "height",
    "width",
    "has_bottom",
    "top",
    "opposite",
    "bottom",
    "lattice",
    "some_not_uppersets",
    "some_lowersets",
    "some_uppersets",
    "some_not_lowersets",
    "some_upper_closures",
    "some_lower_closures",
    "some_maximal",
    "some_supremum",
    "some_upper_bounds",
    "some_minimal",
    "some_lower_bounds",
    "some_infimum",
}
ALLOWED_REQUIRES = {
    "set_product",
    "poset_product",
    "set_union",
    "poset_sum",
}

import os

ENV_VAR = "ACT4E_DATA"


def find_yamls(dirnames: List[str]) -> List[str]:
    res = []
    for dirname in dirnames:
        for f in os.listdir(dirname):
            if f.endswith(".yaml"):
                res.append(os.path.join(dirname, f))
    return res


def get_default_data_dir() -> str:
    dirname = os.path.join(os.path.dirname(__file__), "thedata")
    return dirname


@lru_cache()
def get_all_test_data(load: bool = True) -> Dict[str, TestData[Any]]:
    from_env = os.environ.get(ENV_VAR, None)
    if from_env:
        dirname = from_env
        logger.info(f"loading data from environment variable {ENV_VAR} = {dirname}")

    else:
        msg = (
            f"Using embedded data. You can use the environment variable {ENV_VAR} to give a different "
            f"directory. "
        )
        logger.info(msg)
        dirname = get_default_data_dir()

    dirnames = [dirname]
    yamls = find_yamls(dirnames)
    if not yamls:
        msg = "I could not find any YAML file."
        raise ZValueError(msg, dirnames=dirnames)

    res: Dict[str, TestData[Any]] = {}

    for fn in yamls:
        with open(fn) as f:
            data = f.read()

        d = yaml.load(data, Loader=RoundTripLoader)
        if not isinstance(d, dict):
            msg = f"Invalid data file {fn}"
            raise ZValueError(msg)
        for k, v in d.items():

            if k in res:
                msg = f"Found duplicate key {k} in {fn}"
                raise ZValueError(msg)

            vv = dict(v)
            tags = vv.pop("tags", {})

            try:
                data = vv.pop("data")
            except KeyError:
                raise ZValueError(k=k, v=v)

            requires = vv.pop("requires", {})
            properties = vv.pop("properties", {})

            if vv:
                msg = "Unknown properties"
                raise ZValueError(msg=msg, v=v, vv=vv)

            extra_requires = set(requires) - set(ALLOWED_REQUIRES)
            if extra_requires:
                msg = f'Extra "requires" for entry {k!r}'
                raise ZValueError(msg, extra_requires=extra_requires, allowed=ALLOWED_REQUIRES)

            extra_properties = set(properties) - set(ALLOWED_PROPERTIES)
            if extra_properties:
                msg = "Extra properties"
                raise ZValueError(msg, extra_properties=extra_properties)

            extra_tags = set(tags) - set(ALLOWED_TAGS)
            if extra_tags:
                msg = "Extra tags"
                raise ZValueError(msg, extra_tags=extra_tags)

            res[k] = TestData(tags=tags, requires=requires, data=data, properties=properties)

    entries = {k: v.data for k, v in res.items()}
    if load:
        for k, v in res.items():
            with add_context(k=k):
                res[k].data = substitute(entries, res[k].data)

    return res


def substitute(entries: Dict[str, object], a: object) -> object:
    if isinstance(a, dict):
        if "load" in a:
            loadit = a["load"]
            if loadit not in entries:
                msg = "Cannot find entry to load"
                raise ZValueError(msg, a=a)
            else:
                return substitute(entries, entries[loadit])
        else:
            return {k: substitute(entries, v) for k, v in a.items()}

    elif isinstance(a, list):
        return [substitute(entries, x) for x in a]
    else:
        return a


def purify_data(a: X) -> X:
    """Strip comments etc."""
    loader = YAML()
    loader.indent(mapping=4, sequence=4, offset=2)
    loader.preserve_quotes = True  # type: ignore
    loader.default_flow_style = False
    i = StringIO()
    loader.dump(a, i)
    s = i.getvalue()
    res = yaml.load(s, Loader=yaml.Loader)
    return res


def get_test_relations() -> Dict[str, TestData[I.FiniteRelation_desc]]:
    return get_test_data("relation")


def get_test_posets() -> Dict[str, TestData[I.FinitePoset_desc]]:
    return get_test_data("poset")


def get_test_sets() -> Dict[str, TestData[I.FiniteSet_desc]]:
    return get_test_data("set")


def get_test_maps() -> Dict[str, TestData[I.FiniteSet_desc]]:
    return get_test_data("map")


def get_test_data(tagname: str) -> Dict[str, TestData[Any]]:
    alldata = get_all_test_data()
    res = {}
    for k, v in alldata.items():

        if v.tags.get(tagname, False):
            res[k] = v
    return res


class IOHelperImp(I.IOHelper):
    def loadfile(self, name: str) -> Dict[str, Any]:
        raise NotImplementedError(name)


Rcov = TypeVar("Rcov", covariant=True)
Rcon = TypeVar("Rcon", contravariant=True)

Xcov = TypeVar("Xcov", covariant=True)
Xcon = TypeVar("Xcon", contravariant=True)


class Loader(Protocol[Xcov, Rcon]):
    def load(self, h: I.IOHelper, data: Rcon) -> Xcov:
        ...


class Saver(Protocol[Xcon, Rcov]):
    def save(self, h: I.IOHelper, ob: Xcon) -> Rcov:
        ...


def dumpit_(tc: TestContext, fsr: Saver[Xcon, Rcov], h: I.IOHelper, ob: Xcon) -> Rcov:
    KN = type(fsr).__name__

    try:
        res = fsr.save(h, ob)
    except Exception as e:
        tc.fail(zh.span(f"{KN}:save() raised an exception"), tb=traceback.format_exc(), ob=ob)
        raise ImplementationFail() from e

    if res is None:
        tc.fail(zh.span(f"{KN}:save() returned None"), ob=ob)
        raise ImplementationFail()
    check_good_output(tc, res)
    tc.raise_if_failures()
    return res


def loadit_(tc: TestContext, loader: Loader[Xcov, Rcon], h: I.IOHelper, data: Rcon, K: type) -> Xcov:
    LN = type(loader).__name__
    try:
        res = loader.load(h, data)
    except NotImplementedError:
        raise
    except I.InvalidFormat as e:
        msg = f"Implementation of {LN}.load() threw InvalidFormat but the format is valid."
        tc.fail(zh.span(msg), data=yaml.dump(data), tb=traceback.format_exc())
        raise ImplementationFail() from e
    except BaseException as e:
        msg = f"Implementation of {LN}.load() threw {type(e).__name__} but the format is valid."
        tc.fail(zh.span(msg), data=yaml.dump(data), tb=traceback.format_exc())
        raise ImplementationFail() from e
    tc.expect_type2(res, K, zh.span(f"Expected that {LN}.load() returns a {K.__name__}"))
    tc.raise_if_failures()
    return res


def check_good_output(tc: TestContext, x: object) -> None:
    OK = (int, float, bool, datetime, dict, list, str)
    if isinstance(x, tuple):
        msg = (
            "You cannot use tuples for the concrete representation because they cannot be serialized in "
            "YAML. Try using lists. (You can use tuples for the internal representation.)"
        )
        tc.fail(zh.span(msg), x=x)
        return
    # noinspection PyTypeChecker
    if not isinstance(x, OK):
        msg = (
            f"In the concrete representation you can use only one of the usable datatypes; you used "
            f"an object of type {type(x).__name__}"
        )
        tc.fail(zh.span(msg), x=x)
        return  # raise ZValueError(msg)
    if isinstance(x, list):
        for _ in x:
            check_good_output(tc, _)
    if isinstance(x, dict):
        for k, v in x.items():
            check_good_output(tc, v)


def filter_reqs(d: Dict[str, TestData[X]], req: str) -> Dict[str, TestData[X]]:
    res = {}
    for k, v in d.items():
        requires = set(v.requires)
        if requires == {req}:
            res[k] = v

    return res



from typing import TypeVar, Union

import zuper_html as zh
from zuper_commons.types import ZValueError
from zuper_html_plus import html_from_object
from zuper_testint import TestContext, TestManagerInterface, TestRef, tfor

import act4e_interfaces as I
from .data import check_good_output, get_test_data, IOHelperImp
from .sets_representation import tm_load_set

E = TypeVar("E")
X = TypeVar("X")


def tm_check_subset(
    tm: TestManagerInterface,
    a: Union[I.FiniteSet[X], TestRef[I.FiniteSet[X]]],
    b: Union[I.FiniteSet[X], TestRef[I.FiniteSet[X]]],
    expect: bool,
    tid: str,
) -> TestRef[None]:
    fsp = tm.impof(I.FiniteSetProperties)
    return tm.addtest(check_subset, fsp, a, b, expect, tid0="subset-" + tid)


def test_notnone(tc: TestContext, x: X) -> X:
    if x is None:
        tc.fail(zh.p("None obtained"))
    return x


@tfor(I.FiniteSetProperties)
def test_FiniteSetProperties(tm: TestManagerInterface) -> None:
    sets = get_test_data("set")
    set1 = tm_load_set(tm, "set1", sets["set1"].data)
    set2 = tm_load_set(tm, "set2", sets["set2"].data)
    tm_check_subset(tm, set1, set1, True, "set1-set1")
    tm_check_subset(tm, set1, set2, True, "set1-set2")
    tm_check_subset(tm, set2, set1, False, "set2-set1")

    # print(yaml.dump(alldata))


def check_subset(
    tc: TestContext, fsp: I.FiniteSetProperties, a: I.FiniteSet[X], b: I.FiniteSet[X], expect: bool
) -> None:
    res = fsp.is_subset(a, b)
    if res != expect:
        msg = zh.p(
            "fis_subset() returned",
            zh.code(str(res)),
            " instead of ",
            zh.code(str(expect)),
        )
        tc.fail(msg, a=set(a.elements()), b=set(b.elements()))


def check_save_all_elements(tc: TestContext, S: I.FiniteSet[X]) -> None:
    h = IOHelperImp()
    for a in S.elements():
        ra = S.save(h, a)
        try:
            check_good_output(tc, ra)
        except ZValueError as e:
            msg = zh.span(f"Invalid serialization", html_from_object(e))
            tc.fail(msg, element=a, representation=ra)
            continue  # raise ImplementationFail(msg, element=a, representation=ra) from e
        a2 = S.load(h, ra)
        if not S.equal(a, a2):
            msg = zh.span("Saving and reloading gives a different object")
            tc.fail(msg, element=a, representation=ra, element_reloaded=a2)

            # raise ImplementationFail(msg, element=a, representation=ra, element_reloaded=a2)



from abc import ABC, abstractmethod
from typing import Any, TypedDict

from .helper import IOHelper
from .natural import FiniteNaturalTransformation

__all__ = ["FiniteNaturalTransformationRepresentation", "FiniteNaturalTransformation_desc"]


class FiniteNaturalTransformation_desc(TypedDict):
    pass


class FiniteNaturalTransformationRepresentation(ABC):
    @abstractmethod
    def load(
        self, h: IOHelper, data: FiniteNaturalTransformation_desc
    ) -> FiniteNaturalTransformation[Any, Any, Any, Any]:
        ...

    @abstractmethod
    def save(
        self, h: IOHelper, f: FiniteNaturalTransformation[Any, Any, Any, Any]
    ) -> FiniteNaturalTransformation_desc:
        ...



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Dict

__all__ = ["IOHelper"]


class IOHelper(ABC):
    @abstractmethod
    def loadfile(self, name: str) -> Dict[str, object]:
        """Load from the filesystem."""



import json

from setuptools import setup

with open("setup.json") as f:
    data = json.load(f)

# setup package
params = dict(
    name=data["package_name"],
    author=data["author"],
    author_email=data["author_email"],
    url=data["url"],
    tests_require=data["tests_require"],
    extras_require={"test": data["tests_require"]},
    install_requires=data["install_requires"],
    package_dir={"": data["srcdir"]},
    packages=data["modules"],
    long_description="",
    version=data["version"],
    entry_points=data["entry_points"],
    zip_safe=data["zip_safe"],
    include_package_data=data["include_package_data"],
)

setup(**params)

# sigil ebaeec93e9da74460961ffce57c36fda



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, List

from typing_extensions import TypedDict

from .helper import IOHelper
from .maps import FiniteMap
from .sets_representation import FiniteSet_desc
from .types import ConcreteRepr

__all__ = [
    "FiniteMap_desc",
    "FiniteMapRepresentation",
]


class FiniteMap_desc(TypedDict):
    source: FiniteSet_desc
    target: FiniteSet_desc
    values: List[List[ConcreteRepr]]


class FiniteMapRepresentation(ABC):
    @abstractmethod
    def load(self, h: IOHelper, s: FiniteMap_desc) -> FiniteMap[Any, Any]:
        ...

    @abstractmethod
    def save(self, h: IOHelper, m: FiniteMap[Any, Any]) -> FiniteMap_desc:
        ...



from abc import ABC, abstractmethod
from typing import Any, TypedDict

from .categories_representation import FiniteCategory_desc
from .functors import FiniteFunctor
from .helper import IOHelper
from .maps_representation import FiniteMap_desc

__all__ = ["FiniteFunctor_desc", "FiniteFunctorRepresentation"]


class FiniteFunctor_desc(TypedDict):
    source: FiniteCategory_desc
    target: FiniteCategory_desc
    f_ob: FiniteMap_desc
    f_mor: FiniteMap_desc


class FiniteFunctorRepresentation(ABC):
    @abstractmethod
    def load(self, h: IOHelper, data: FiniteFunctor_desc) -> FiniteFunctor[Any, Any, Any, Any]:
        ...

    @abstractmethod
    def save(self, h: IOHelper, f: FiniteFunctor[Any, Any, Any, Any]) -> FiniteFunctor_desc:
        ...
